"""
Module for generic states and state ABCs.

Custom states may be defined for etiher Operational or Action states, but should follow the convention for full
functionality.

# Operational states:

- an offline or unavailable state should be indicated by 0
- operational states should be indicated by values 1 or greater
- inoperable or error states should be indicated by values -1 or less

# Action states

- use 0 to represent an inactive state
- use values 1 or greater to indicate active states


# Custom states

Custom states are supported where a enumeration value is an instance of a State subclass. This adds support for
callbacks when a state is switched to.
"""

import logging
import warnings
from enum import Enum
from typing import Callable, Union, List, Optional

from ..action.configured import ConfiguredAction, ConfiguredActionList, TrackedAction


state_logger = logging.getLogger('hein_control.states')
state_logger.setLevel(logging.INFO)


class State:
    def __init__(self,
                 value: int,
                 *callbacks: Callable,
                 alias: str = None,
                 meaning: str = None,
                 ):
        """
        Custom state definition with additional support. An integer value must be specified, and callbacks are supported.
        If callbacks are specified for a state, these callbacks will be automatically executed when the state is set.

        :param value: state value
        :param callbacks: callback functions. These functions should take no arguments.
        :param alias: name alias for convenience
        :param meaning: optional meaning for the state (i.e. a description of what the state is and what it implies)
        """
        self.value = value
        self.alias = alias
        self.meaning = meaning
        self.pre_state_callbacks: ConfiguredActionList = ConfiguredActionList(use_unique_action=True)
        self.register_pre_state_callbacks(*callbacks)
        self.post_state_callbacks = ConfiguredActionList(use_unique_action=True)

    def __str__(self):
        return f'{self.value} ({len(self.pre_state_callbacks) + len(self.post_state_callbacks)} callbacks)'

    def __eq__(self, other: Union[int, 'State']):
        if isinstance(other, State):
            return other.value == self.value
        return other == self.value

    def __lt__(self, other: Union[int, 'State']):
        if isinstance(other, State):
            return self.value < other.value
        return self.value < other

    def __gt__(self, other: Union[int, 'State']):
        if isinstance(other, State):
            return self.value > other.value
        return self.value > other

    def __le__(self, other: Union[int, 'State']):
        return self < other or self == other

    def __ge__(self, other: Union[int, 'State']):
        return self > other or self == other

    @property
    def has_pre_state_change_callbacks(self) -> bool:
        """whether the State has defined pre-state-change callbacks"""
        return len(self.pre_state_callbacks) > 0

    @property
    def has_post_state_change_callbacks(self) -> bool:
        """whether the State has defined post-state-change callbacks"""
        return len(self.post_state_callbacks) > 0

    @property
    def has_callbacks(self) -> bool:
        """whether the State instance has registered callbacks"""
        return self.has_pre_state_change_callbacks or self.has_post_state_change_callbacks

    def register_pre_state_callbacks(self, *callbacks: Callable):
        """
        Registers the provided callbacks with the State instance. These callbacks will be executed before a Component
        state manager switches to this state.

        :param callbacks: callbacks to register
        """
        return self.register_callbacks(*callbacks, pre_state=True)

    def register_post_state_callbacks(self, *callbacks):
        """
        Registers the provided callbacks with the State instance. These callbacks will be executed after a Component
        state manager switches to this state.

        :param callbacks: callbacks to register
        """
        return self.register_callbacks(*callbacks, pre_state=False)

    def register_callbacks(self,
                           *callbacks: Callable,
                           pre_state: bool = True,
                           parent_id: Optional[Union[bool, str]] = None
                           ):
        """
        Registers the provided callbacks with the State instance. These callbacks will be executed when a Component
        state manager switches to this state.

        :param callbacks: callbacks to register
        :param pre_state: whether the callbacks should be pre or post state transition
        :param parent_id: optional parent ID override for created ConfiguredAction instance (e.g. associating a callback
            with a specific state)
        """
        callback_list = self.pre_state_callbacks if pre_state is True else self.post_state_callbacks
        for callback in callbacks:
            if callable(callback) is False:
                raise TypeError(f'the provided callback is not callable: {callback}')
            if callback_list.action_registered_in_list(callback) is False:
                state_logger.debug(f'registering {"pre" if pre_state else "post"}-state change {callback} to {self}')
                callback = ConfiguredAction(
                    callback,
                    parent_identifier_override=parent_id or callback_list.parent_identifier_override,
                )
                callback_list.append(callback)

    def deregister_callbacks(self, *callbacks: Callable, remove_configurations: bool = False):
        """
        De-registers callbacks from the State. If a callback is not registered with the instance that callbck will be
        ignored.

        :param callbacks: callbacks to remove from the instance
        :param remove_configurations: option to remove ConfiguredAction from registry after removal from the list
            (useful for one-off registered methods that will never be used again)
        """
        # todo catch the event that the same callback is registered both before and after the event
        for callback in callbacks:
            for callback_list in [self.pre_state_callbacks, self.post_state_callbacks]:
                if callback_list.action_registered_in_list(callback) is True:
                    ca = callback_list.get_ca_by_action(callback)
                    state_logger.debug(f'removing {ca.name} from {self} callbacks')
                    callback_list.remove(ca)
                    if remove_configurations is True:
                        ConfiguredAction.remove_configuration_from_registry(ca)

    def execute_callbacks(self) -> List:
        """executes any registered callbacks and returns their return values in order"""
        warnings.warn(  # v7.1.9
            'the method has been deprecated, call execute_pre_state_callbacks or execute_post_state_callbacks directly',
            DeprecationWarning,
            stacklevel=2,
        )
        return self.execute_pre_state_callbacks()

    def execute_pre_state_callbacks(self) -> List[TrackedAction]:
        """executes any registered pre-state-change callbacks and returns their values in order"""
        out = []
        for callback in self.pre_state_callbacks:
            ta = callback.get_tracked_from_config()
            ta()
            out.append(ta)
        return out

    def execute_post_state_callbacks(self) -> List[TrackedAction]:
        """executes any registered post-state-change callbacks and returns their values in order"""
        out = []
        for callback in self.post_state_callbacks:
            ta = callback.get_tracked_from_config()
            ta()
            out.append(ta)
        return out


class StateSet(Enum):
    """A pseudo-IntEnum abstract base class which also supports State instances as values"""

    @classmethod
    def _missing_(cls, value):
        """missing hook for the Enum class which supports the State class"""
        for name, state in cls._member_map_.items():
            # match value to state value
            if value == state.value:
                return state
            elif value == state.get_name():
                return state
            elif value == state.name:
                return state
        raise ValueError(f'{value} is not a valid {cls.__name__}')

    def __lt__(self, other: Union[int, 'StateSet']):
        if isinstance(other, StateSet):
            return self.value < other.value
        return self.value < other

    def __gt__(self, other: Union[int, 'StateSet']):
        if isinstance(other, StateSet):
            return self.value > other.value
        return self.value > other

    def __eq__(self, other: Union[int, 'StateSet']):
        if isinstance(other, StateSet):
            return self.value == other.value
        return self.value == other

    def __le__(self, other: Union[int, 'StateSet']):
        return self < other or self == other

    def __ge__(self, other: Union[int, 'StateSet']):
        return self > other or self == other

    @property
    def has_pre_state_change_callbacks(self) -> bool:
        """whether the value has defined pre-state-change callbacks"""
        if isinstance(self.value, State):
            return self.value.has_pre_state_change_callbacks
        return False

    @property
    def has_post_state_change_callbacks(self) -> bool:
        """whether the value has defined post-state-change callbacks"""
        if isinstance(self.value, State):
            return self.value.has_post_state_change_callbacks
        return False

    @property
    def has_callbacks(self) -> bool:
        """whether the value has callbacks registered"""
        return self.has_post_state_change_callbacks or self.has_pre_state_change_callbacks

    def get_name(self) -> str:
        """retrieves the name or alias of the state"""
        if isinstance(self.value, State):
            return self.value.alias or self._name_
        return self.name

    def set_alias(self, alias: str = None):
        """sets the alias for the StateSet instance"""
        if isinstance(self.value, State) is False:
            raise TypeError(f'the {self.__class__.__name__} value is not a State instance')
        self.value.alias = alias

    def register_callbacks(self, *callbacks: Callable, pre_state: bool = True):
        """
        Pass through for registering callbacks with the associated State instance. These callbacks will be executed
        when a Component state manager switches to this state.

        :param callbacks: callbacks to register
        :param pre_state: whether the callbacks should be before or after the state change
        """
        if isinstance(self.value, State) is False:
            raise TypeError(f'the {self.__class__.__name__} value is not a State instance')
        self.value.register_callbacks(*callbacks, pre_state=pre_state)

    def register_pre_state_callbacks(self, *callbacks):
        """
        Pass through for registering callbacks with the associated State instance. These callbacks will be executed
        before a Component state manager switches to this state.

        :param callbacks: callbacks to register
        """
        return self.register_callbacks(*callbacks, pre_state=True)

    def register_post_state_callbacks(self, *callbacks):
        """
        Pass through for registering callbacks with the associated State instance. These callbacks will be executed
        after a Component state manager switches to this state.

        :param callbacks: callbacks to register
        """
        return self.register_callbacks(*callbacks, pre_state=False)

    def deregister_callbacks(self, *callbacks: Callable, remove_configurations: bool = False):
        """
        Pass through for de-registering callbacks with the associated State instance.

        :param callbacks: callbacks to remove from the instance
        :param remove_configurations: option to remove ConfiguredAction from registry after removal from the list
            (useful for one-off registered methods that will never be used again)
        """
        if isinstance(self.value, State) is False:
            raise TypeError(f'the {self.__class__.__name__} value is not a State instance')
        self.value.deregister_callbacks(*callbacks, remove_configurations=remove_configurations)

    def execute_callbacks(self) -> List:
        """
        executes any registered callbacks and returns their return values in order. If the value is not a State
        instance, no action will be taken
        """
        warnings.warn(  # v7.1.9
            'call execute_pre_state_callbacks or execute_post_state_callbacks directly',
            DeprecationWarning,
            stacklevel=2,
        )
        return self.execute_pre_state_callbacks()

    def execute_pre_state_callbacks(self) -> List:
        """
        executes any registered pre-state-change callbacks and returns their return values in order. If the value is
        not a State instance, no action will be taken
        """
        if isinstance(self.value, State):
            return self.value.execute_pre_state_callbacks()
        return []

    def execute_post_state_callbacks(self) -> List:
        """
        executes any registered post-state-change callbacks and returns their return values in order. If the value is
        not a State instance, no action will be taken
        """
        if isinstance(self.value, State):
            return self.value.execute_post_state_callbacks()
        return []


class OperationalState(StateSet):
    """
    operational state tracker for components. By convention, use 0 to indicate offline, negative values for
    error states, and positive values for online states
    """
    ERROR = State(-1)
    OFFLINE = State(0)
    ONLINE = State(1)


class GenericActionState(StateSet):
    """
    Basic action state tracker for component. By convention, use 0 to represent an inactive state for the inactive
    state check decorator to function.
    """
    IDLE = State(0)
    ACTIVE = State(1)






"""mixin classes for use across the package"""

from .status import Status
from .config import Configuration
from .reg import InstanceRegistry, ClassProperty
from .serialization import (DumpableState, DumpableList, is_jsonable, make_iterable_serializable,
                            make_mapping_serializable, serializable_cast)

__all__ = [
    'Status',
    'Configuration',
    'InstanceRegistry',
    'ClassProperty',
    'DumpableState',
    'DumpableList',
    'is_jsonable',
    'make_iterable_serializable',
    'make_mapping_serializable',
    'serializable_cast',
]



"""
Module for interacting with an SQLite3 database (.db files)
"""
import sqlite3
import pathlib

from abc import ABC, abstractmethod
from typing import List, Tuple, Union, Dict
from contextlib import closing
from datetime import datetime, timedelta

import warnings


class MissingColumnError(Exception):
    """
    Error for a missing column in a DatabaseTable
    """

    def __init__(self, msg=None):
        self.msg = msg or 'An expected column is missing from the database table'

    def __str__(self):
        return self.msg


def get_db_table_names(file_path: str) -> List[str]:
    """Return a list of table names, given the database path"""
    with sqlite3.connect(file_path) as db:
        with closing(db.cursor()) as cur:
            cur.execute(f"SELECT name FROM sqlite_master WHERE type='table'")
            table_names = [names[0] for names in cur.fetchall()]
    return table_names


def get_db_column_names(file_path, table_name: str) -> List[str]:
    """Return a list of the column names for a table in a database, given the database path"""
    raise NotImplementedError('access the get_table_column_names method of DatabaseTable')


class DatabaseTable(ABC):
    def __init__(self, file_path: Union[str, pathlib.Path]):
        """
        Create a SQLite3 table at the database at the path with the columns if it doesnt already exist

        This is a SQLite3 database table mixin. To use, subclass this class and define the table_name and
        column_name_types. Basic sql injection attack prevention is provided by disallowing the user to set the
        table name and column definitions (defined when the class is defined). There is no protection against imporper
        definition of the subclasses.

        :param file_path: path to the database (.db file)
        """
        self._file_path: pathlib.Path = None
        self.file_path = file_path

        # create table if it does not exist
        commands = []
        # build command and parameters so that they can be executed safely
        for name, value_type in self.column_name_types:
            commands.append(f'{name} {value_type}')
        table_structure = ', '.join(commands)
        with sqlite3.connect(self.file_path) as db:
            with closing(db.cursor()) as cur:
                # create the table
                cur.execute(
                    f'CREATE TABLE IF NOT EXISTS {self.table_name} ({table_structure})',
                )
            db.commit()

        self.column_names: List[str] = self.get_table_column_names()

    @property
    def file_path(self) -> pathlib.Path:
        """Path to the database .db file"""
        return self._file_path

    @file_path.setter
    def file_path(self, value):
        if value is not None:
            if isinstance(value, pathlib.Path) is False:
                value = pathlib.Path(value)
            if value.suffix != '.db':
                value = value.parent / (value.name + '.db')
            self._file_path = value

    @property
    @abstractmethod
    def table_name(self) -> str:
        """Name of the table in the database"""
        raise NotImplementedError

    @property
    @abstractmethod
    def column_name_types(self) -> List[Tuple[str, str]]:
        """defines the column names and types to be used if the table needs to be created"""
        raise NotImplementedError

    @property
    def _base_query(self) -> str:
        """the base query for the instance. This provides a basic security layer between the user and the query for
        avoiding sql injections"""
        warnings.warn('use _select_query instead', DeprecationWarning, stacklevel=2)
        return self._select_query

    @property
    def _select_query(self) -> str:
        """the select query for the instance. This provides a basic security layer between the user and the query for
        avoiding sql injections"""
        return f'select * from {self.table_name}'

    @property
    def _insert_query(self) -> str:
        """the delete query for the instance. This provides a basic security layer between the user and the
        query for avoiding sql injections"""
        return f'insert into {self.table_name} values ({", ".join(["?" for _ in range(len(self.column_names))])})'

    @property
    def _delete_query(self) -> str:
        """the base query for the instance. This provides a basic security layer between the user and the query for
        avoiding sql injections"""
        # todo should it be that this must be implemented in the subclass to for safety?
        return f'delete from {self.table_name}'

    @property
    def db_table_names(self) -> List[str]:
        """A list of the name of all tables in the database"""
        names = get_db_table_names(file_path=self.file_path)
        return names

    def get_table_column_names(self) -> List[str]:
        """Return a list of the column names for a table in a database, given the database path"""
        with sqlite3.connect(self.file_path) as db:
            with closing(db.cursor()) as cur:
                # todo scrub for sql injection
                cur.execute(self._select_query)
                column_names = [description[0] for description in cur.description]
        return column_names

    def insert(self, values) -> None:
        """
        Inserts values into the database table. The order of the values in the Tuple must match to how they are
        listed in the column_name_types property

        :param values: tuple of the values to insert into the database table
        :return:
        """
        with sqlite3.connect(self.file_path) as db:
            with closing(db.cursor()) as cur:
                cur.execute(self._insert_query, values)
                db.commit()


class TimeBasedDatabaseTable(DatabaseTable):
    def __init__(self, file_path: Union[str, pathlib.Path]):
        """
        A DatabaseTable, with additional methods to retrieve and delete entries based on time; a TimeBasedDatabaseTable
        MUST have a column called 'datetime'

        This is a SQLite3 database table mixin. To use, subclass this class and define the table_name and
        column_name_types. Basic sql injection attack prevention is provided by disallowing the user to set the
        table name and column definitions (defined when the class is defined). There is no protection against imporper
        definition of the subclasses. All subclasses MUST have a column called 'datetime'

        :param file_path: path to the database (.db file)
        """
        column_names = [name for name, value_type in self.column_name_types]
        if 'datetime' not in column_names:
            raise MissingColumnError('datetime must be a column in the table')
        DatabaseTable.__init__(self, file_path)

    @abstractmethod
    def retrieve(self,
                 start_time: datetime = None,
                 end_time: datetime = None,
                 duration: timedelta = timedelta(hours=1),
                 ) -> List[List]:
        """
        Retrieve entries that fall within a time range (inclusive of the start and end times). Must be overwritten
        by the subclass, and the subclass must have a column called "datetime".

        Either provide:
            - start_time and end_time, then duration won't be used
            - start_time and duration, then end_time is start_time + duration
            - end_time and duration, then start_time is end_time - duration
            - only duration, then a default end_time of now is used. if duration is None, then return all entries in
                the database disregarding a start or end time

        Example of how this might be implemented:
            if duration is not None:
                if start_time is None and end_time is None:
                    end_time = datetime.now()

                if start_time and end_time:
                    pass
                elif start_time:
                    end_time = start_time + duration
                elif end_time:
                    start_time = end_time - duration
                command = self._select_query + " WHERE datetime BETWEEN :start_time AND :end_time"
                values = {
                    'start_time': start_time,
                    'end_time': end_time,
                }
            else:
                command = self._select_query
                values = {}
            with sqlite3.connect(self.file_path) as db:
                with closing(db.cursor()) as cur:
                    cur.execute(command, values)
                    rows = cur.fetchall()
            out = []
            for row in rows:
                # cast each value to the types they should be
                try:
                    dt = datetime.strptime(row[0], "%Y-%m-%d %H:%M:%S.%f")
                except ValueError:
                    dt = datetime.strptime(row[0], "%Y-%m-%d %H:%M:%S")
                # overwrite until out.append(r) to properly cast the values for the subclass and add them to r
                col_1 = str(row[1])
                col_2 = int(row[2])
                col_3 = bool(row[3])
                # add the row of values to the output
                r = [dt, col_1, col_2, col_3]
                out.append(r)
            return out

        :param start_time: datetime object to denote the start of the time range to retrieve entries
        :param end_time: datetime object to denote the end of the time range to retrieve entries
        :param duration: timedelta object for how long the the time range to retrieves entries is. Default 1 hour
        :return: a list of lists. the inner lists each represent a row from the database, and each value in the inner
            list represents the value for a column index
        """
        raise NotImplementedError

    def retain(self,
               start_time: datetime = None,
               end_time: datetime = None,
               duration: timedelta = timedelta(days=31),
               ) -> None:
        """
        Only retain the entries that fall within a time range and delete everything else from the database table
        (retain entries that fall on the start and end times).

        'datetime' must be a column in the database table

        Either provide:
            - start_time and end_time, then duration won't be used
            - start_time and duration, then end_time is start_time + duration
            - end_time and duration, then start_time is end_time - duration
            - only duration, then a default end_time of now is used

        https://sqlite.org/lang_delete.html

        :param start_time: datetime object to denote the start of the time range to retain entries
        :param end_time: datetime object to denote the end of the time range to retain entries
        :param duration: timedelta object for how long the the time range to retain entries is. Default 31 days
        :return:
        """
        if 'datetime' not in self.column_names:
            raise MissingColumnError('datetime must be one of the columns in the database')

        if start_time is None and end_time is None:
            end_time = datetime.now()

        if start_time and end_time:
            pass
        elif start_time:
            end_time = start_time + duration
        elif end_time:
            start_time = end_time - duration

        command = self._delete_query + " WHERE datetime NOT BETWEEN :start_time AND :end_time"
        values = {
            'start_time': start_time,
            'end_time': end_time,
        }

        with sqlite3.connect(self.file_path) as db:
            with closing(db.cursor()) as cur:
                cur.execute(command, values)


class DatabaseRetentionPolicy(ABC):
    def __init__(self, enabled: bool = False):
        """
        A retention policy that can be run by on a DatabaseTable; it should remove entries from the .db SQLite database
        based on the _retention_policy

        This is an abstract base class. To use, subclass this class and define the _retention_policy. There is no
        protection against improper definition of the subclasses. Since an exact _retention_policy must be defined,
        it is up to the implementor to ensure that the retention policy will successfully execute on the provided
        DatabaseTable instance

        :param enabled: whether to actually execute the retention policy on a DatabaseTable when a
            DatabaseRetentionPolicy instance is called
        """
        self._enabled = enabled

    @property
    @abstractmethod
    def _retention_policy(self) -> str:
        """The exact SQL statement to execute to edit a SQLite .db database table"""
        raise NotImplementedError

    def __call__(self, *args, **kwargs) -> bool:
        """Execute the retention policy for the database if the policy is enabled and return whether the retention
        policy was actually run or not"""
        return self._run(*args, **kwargs)

    @property
    @abstractmethod
    def _retention_policy_values(self) -> Dict:
        """Dictionary for the _retention_policy. Keys should match the variables that should be replaced in the
        _retention_policy"""
        raise NotImplementedError

    @property
    def enabled(self) -> bool:
        """Whether the retention policy is enabled and should actually be executed when the DatabaseRetentionPolicy
        instance is called"""
        return self._enabled

    @enabled.setter
    def enabled(self, value: bool) -> None:
        if value:
            self._enabled = value

    def _run(self, database_table: DatabaseTable) -> bool:
        """
        Execute the retention policy for the database if the policy is enabled

        :param database_table: the DatabaseTable that will run the policy
        :return: bool, whether the retention policy was run or not
        """
        if self.enabled:
            with sqlite3.connect(database_table.file_path) as db:
                with closing(db.cursor()) as cur:
                    cur.execute(self._retention_policy, self._retention_policy_values)
            return True
        else:
            return False



"""mixins for ensuring/verifying JSON serializability"""

import json
from abc import ABC, abstractmethod
from typing import Union, Any


def is_jsonable(obj) -> bool:
    """verifies that the provided object is JSON serializable"""
    try:
        json.dumps(obj)
        return True
    except (TypeError, OverflowError):
        return False


class DumpableState(ABC):
    """a dumpable object which can be represented as a dictionary"""

    @abstractmethod
    def as_dict(self) -> dict:
        """represents the state or properties as a serializable object"""
        raise NotImplementedError


class DumpableList(ABC):
    """a dumpable object which can be represented as a list"""

    @abstractmethod
    def as_list(self) -> list:
        """if the dumpable is iterable, represents the list as a list"""
        raise NotImplementedError


def serializable_cast(incoming: Union[DumpableState, DumpableList]) -> Union[Any, list, dict]:
    """attempts to cast a DumpableState or DumpableList to serializable"""
    if is_jsonable(incoming):
        return incoming
    elif isinstance(incoming, DumpableList):
        return incoming.as_list()
    elif isinstance(incoming, DumpableState):
        return incoming.as_dict()
    else:
        raise TypeError(f'the provided object type "{type(incoming)}" could not be made serializable')


def make_iterable_serializable(*obj: Union[DumpableState, DumpableList, Any]) -> list:
    """
    Attempts to ensure that the provided object(s) are serializable.

    :param obj: objects to ensure are serializable
    :return: serializable object
    """
    out = []
    for o in obj:
        # if it's serializable, directly append
        if is_jsonable(o) is True:
            out.append(o)
        else:  # otherwise attempt cast
            out.append(serializable_cast(o))
    return out


def make_mapping_serializable(**objs) -> dict:
    """
    Attempts to ensure that the provided object(s) are serializable

    :param objs: key-object mappings
    :return: serializable dictionary
    """
    out = {}
    for key, obj in objs.items():
        if is_jsonable(obj):
            out[key] = obj
        else:
            out[key] = serializable_cast(obj)
    return out



"""manages registry of class instances across the package"""
import uuid
import logging
from typing import MutableMapping, List


logger = logging.getLogger(__name__)


class ClassProperty(object):
    def __init__(self, f):
        """
        Decorator for indicating a class property. Only works for getters
        lovingly pilfered from https://stackoverflow.com/a/5192374

        :param f: class function to decorate
        """
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)


class InstanceRegistry:
    # registry dictionary for all instances of all subclasses
    _registry: MutableMapping[str, List['InstanceRegistry']] = {}

    # prefix for UUID access
    UUID_PREFIX: str = ''

    def __init__(self):
        """A class for tracking class instances and retrieving them by UUID"""
        # todo create unittests
        self._uuid: str = str(uuid.uuid4())   # create uuid
        # append instance to registry
        # logger.debug(f'appending instance with uuid {self._uuid}')
        if len(self.class_registry) == 0 or not any([self is inst for inst in self.class_registry]):
            self.class_registry.append(self)
            # self._registry[f'{self.__class__.__name__}'].append(self)

    @property
    def uuid(self) -> str:
        """unique identifier for the instance"""
        return self.UUID_PREFIX + self._uuid

    @ClassProperty
    def class_registry(cls):
        """a list of registered instances of the current class"""
        # create if undefined
        if cls.__name__ not in cls._registry:
            logger.debug(f'creating new registry list for {cls.__name__}')
            cls._registry[cls.__name__] = []
        return cls._registry[cls.__name__]

    @staticmethod
    def _match_instance_key(inst: "InstanceRegistry", identifier: str) -> bool:
        """
        Checks the provided identifier key against the instance provided, returning True if matching and False if not

        :param inst: InstanceRegistry instance
        :param identifier: string identifier
        :return: match
        """
        if inst._uuid == identifier:
            return True
        elif inst.uuid == identifier:
            return True
        # todo consider validating against concatenated versions of the uuid
        # todo consider breaking out into a matching class (to add matches easily without overwriting this function)
        return False

    @classmethod
    def class_instance_by_id(cls, identifier: str):
        """
        Retrieves the instance of the current class by UUID

        :param identifier: unique identifier
        :raises NameError: if no matching instance could be found
        :return: registered class instance
        """
        try:
            for inst in cls._registry[cls.__name__]:
                if cls._match_instance_key(inst, identifier) is True:
                    return inst
        except KeyError:  # if there are no registered instances of the class, raise a consistent error
            pass
        raise NameError(f'the uuid "{identifier}" could not be found in the registry of {cls.__name__}')

    @classmethod
    def search_class_for_id(cls, identifier: str, class_name: str):
        """
        Searches the defined class for the uuid provided

        :param identifier: string identifier
        :param class_name: class name to search
        :return: instance
        """
        try:
            for inst in cls._registry[class_name]:
                if cls._match_instance_key(inst, identifier) is True:
                    return inst
        except KeyError:  # if there are no registered instances of the class, raise a consistent error
            pass
        except AttributeError:  # if the match instance searches for undefined attributes, raise a consistent error
            pass
        raise NameError(f'the uuid "{identifier}" could not be found in the registry of {class_name}')

    @classmethod
    def get_instance_by_uuid(cls, identifier: str):
        """
        General retrieval of any registered instance by its uuid.

        :param identifier: string identifier
        :return: instance
        """
        for cls_name in cls._registry:
            try:
                return cls.search_class_for_id(identifier, cls_name)
            except NameError:
                continue
        raise NameError(f'the uuid "{identifier}" could not be matched to any instances in the registry')



"""
Module for managing time points in a sequence. `TimePoint` simply manages a time point relative to some start,
and `ActionTimePoint` manages context, sequences of actions, and execution of those actions when desired.
"""
import datetime
import uuid
import logging
import warnings
import time

from typing import Union, Callable, Iterable
from .mixin import InstanceRegistry, DumpableState
from .action import Action, ConfiguredAction, ConfiguredActionList
from .sequencing import ExecutionSequence, ConfiguredSequence
from .action.configured import ensure_triggered

logger = logging.getLogger(__name__)


class TimePoint(InstanceRegistry, DumpableState):
    UUID_PREFIX = 'tp-'

    def __init__(self,
                 time_delta: float,
                 name: str = None,
                 description: str = None,
                 parent_sequence_id: str = None,
                 ):
        """
        A class for describing and tracking an event time point relative to some start time.

        :param time_delta: time delta for the time point (seconds; relative to some other time)
        :param name: convenience name for time point differentiation
        :param description: extended description for differentiation
        :param parent_sequence_id: uuid of parent SamplingScheduler (if applicable)
        """
        InstanceRegistry.__init__(self)
        # time delta from trigger time
        self.time_delta: float = float(time_delta)
        # storage for start time
        self.sequence_time_started: float = None
        # name for convenience
        self.name: str = name
        # description
        self.description: str = description
        # parent uuid
        self.parent_sequence_id: str = parent_sequence_id

        self._uuid = str(uuid.uuid4())  # assign unique identifier
        # todo add description attribute
        # todo create association between time point and parent scheduler
        # todo finish refactor
        #   - move as_dict, update as necessary

    def __str__(self):
        return (
            f'{self.__class__.__name__} {self.time_delta} s '
        )

    def __repr__(self):
        return self.__str__()

    @property
    def uuid(self) -> str:
        """uuid for the time point"""
        return self.UUID_PREFIX + self._uuid

    @classmethod
    def get_instance(cls, key: str) -> "TimePoint":
        """
        Retreives a timepoint instance by key. The key may be a uuid, or class-uuid

        :param key: key to search by
        :return: TimePoint (or subclass) instance
        """
        warnings.warn(
            'the get_instance method has been deprecated, use class_instance_by_id',
            DeprecationWarning,
            stacklevel=2,
        )
        return cls.class_instance_by_id(key)

    def as_dict(self) -> dict:
        """Returns a key: value representation of the instance"""
        dct = super(TimePoint, self).as_dict()
        dct.update({
            'time_delta': self.time_delta,
            'uuid': self.uuid,
            'name': self.name,
            'class': self.__class__.__name__,
        })
        return dct

    def sample_time_from_reference(self, reference_time: float):
        """
        Returns the sample time for the time point from the reference time (in seconds)

        :param reference_time: reference time to calculate from
        :return: time to trigger (seconds)
        """
        return self.time_delta + reference_time

    def __lt__(self, other: Union[float, "TimePoint"]):
        if isinstance(other, TimePoint):
            other = other.time_delta
        return self.time_delta < other

    def __eq__(self, other: Union[float, "TimePoint"]):
        if isinstance(other, TimePoint):
            other = other.time_delta
        return self.time_delta == other

    def __gt__(self, other: Union[float, "TimePoint"]):
        if isinstance(other, TimePoint):
            other = other.time_delta
        return self.time_delta > other

    def __le__(self, other: Union[float, "TimePoint"]):
        return self.__lt__(other) or self.__eq__(other)

    def __ge__(self, other: Union[float, "TimePoint"]):
        return self.__gt__(other) or self.__eq__(other)


class ActionTimePoint(TimePoint, ExecutionSequence):
    UUID_PREFIX = 'atp-'

    def __init__(self,
                 time_delta: float,
                 configuration: Union[str, ConfiguredSequence],
                 name: str = None,
                 **action_kwargs,
                 ):
        """
        A class for describing and tracking an event time point relative to some other time where actions
        (e.g. a sampling sequence) are to be triggered. The sequence of actions is stored in a ConfiguredSequence
        instance.

        :param time_delta: time delta for the time point (relative to some other time)
        :param name: convenience name for time point differentiation
        :param actions: action methods to call. These methods will be called in the order provided
        :param action_kwargs: action kwargs to associate with list of actions (primary keys must map to function names)
        """
        # todo create abstract class with context
        #   - individual methods for pre-step, main, and post-step (e.g. turn on pump, run experiment, process data)
        # todo associate timepoint with parent scheduler instance (by UUID)
        # todo create passthrough for sequence_time_started on init
        TimePoint.__init__(
            self,
            time_delta=time_delta,
            name=name,
        )
        ExecutionSequence.__init__(
            self,
            configuration=configuration,
            **action_kwargs,
        )

    def __str__(self):
        out = f'{self.__class__.__name__} {self.string_time_delta}'
        if self.complete is True:
            out += ' COMPLETE'
        elif self.triggered is True:
            out += ' TRIGGERED'
        return out

    def _executor(self, *args, **kwargs):
        """
        Target method for the internal thread.
        The ThreadedTrackedAction _executor method is leveraged with the start and completion points noted in the log.
        """
        logger.info(f'{self} sequence starting')
        super(ActionTimePoint, self)._executor(*args, **kwargs)
        logger.info(f'{self} sequence completed, duration: {self.action_duration}')

    def remove_action(self,
                      action: Union[str, Callable, Action],
                      ):
        """
        Removes an action from the list of actions associated with a timepoint.

        :param action: action identifier
        :return:
        """
        raise NotImplementedError()
        # todo implement
        #   - determine identifier
        #   - remove from _actions list

    def trigger(self,
                wait: bool = None,
                **kwargs,
                ):
        """
        Triggers (starts execution of) the action sequence of the ActionTimePoint.
        Keyword arguments will be passed through to their respective action methods by the name of the action.
        This method executes the built-in __call__ of ExecutionSequence and it will block until the thread of the
        ExecutionSequence has started.

        e.g. if there is a method named "do_this_thing" which expects the keyword argument "value", the expected syntax
        of kwargs would be

            {
                "do_this_thing": {"value": 42},
            }

        Execution Tree:
        1. Execution sequence __call__
        2. ThreadedTrackedAction using operators.execute_sequence_actions action
        3. Waits until the thread starts (prevents double-triggering
        4. waits for completion if specified


        :param wait: whether to wait for completion before returning
        :param kwargs: keyword arguments for the actions. Keyword arguments will be passed through to their respective
            methods.

        """
        # todo refactor to trigger_and_wait (separate so that signatures match)
        self.__call__(**kwargs)
        # block until triggered flag is set (prevents double-triggering in the same loop)
        while self._triggered is False:
            time.sleep(0.001)
        if wait is True:
            self.wait_for_completion()

    @property
    def string_time_delta(self) -> str:
        """formatted string version of time delta"""
        return str(
            datetime.timedelta(seconds=self.time_delta)
        )

    @property
    @ensure_triggered
    def relative_start_times(self) -> dict:
        """relative start times for actions"""
        if self.sequence_time_started is None:
            raise ValueError(f'sequence_start_time is not set for the {self.__class__.__name__} instance')
        return {
            action.configuration.name: action.time_started - self.sequence_time_started
            for action in self.args
            if action.time_started is not None
        }

    def copy_configuration(self) -> "ActionTimePoint":
        """
        Copies the configuration of an ActionTimePoint instance.

        :return: ActionTimePoint instance with identical configuration
        """
        out = self.__class__(
            time_delta=self.time_delta,
            name=self.name,
            configuration=self.configuration,
        )
        out.update_kwargs(**self.kwargs)
        return out

    @classmethod
    def create_from_actions(cls,
                            time_delta: float,
                            actions: Iterable[Union[str, 'ConfiguredAction']],
                            name: str = None,
                            ) -> "ActionTimePoint":
        """
        Creates an instance from a list of configured actions. This differs from calling init by accepting
        an iterable argument containing actions which are used to create the ConfiguredSequence required by
        init.

        :param time_delta: time delta for the time point (relative to some other time)
        :param actions: configured actions or named configurations
        :param name: name for the time point (optional)
        :return: configured instance
        """
        config = ConfiguredSequence(
            *ConfiguredActionList(*actions)
        )
        return cls(
            time_delta=time_delta,
            configuration=config,
            name=name,
        )

    def as_dict(self) -> dict:
        """Returns a key: value representation of the instance"""
        # todo decide how to include relative start times
        return super(ActionTimePoint, self).as_dict()




"""module for managing component states"""

from .abc import OperationalState, GenericActionState, State, StateSet
from .component import (BadComponentState, ensure_component_online, ensure_component_inactive, ComponentState,
                        TemporaryOperationalCallback, TemporaryActionCallback, TemporaryCallbackSetter)


__all__ = [
    'State',
    'StateSet',
    'BadComponentState',
    'ensure_component_inactive',
    'ensure_component_online',
    'ComponentState',
    'OperationalState',
    'GenericActionState',
    'TemporaryActionCallback',
    'TemporaryCallbackSetter',
    'TemporaryOperationalCallback',
]



"""
Tests database tools
"""
import pathlib
import time
import unittest
from pathlib import Path
import threading
from typing import List, Tuple, Dict

import sqlite3
from contextlib import closing
from datetime import datetime, timedelta

from hein_control.db_tools import DatabaseTable, TimeBasedDatabaseTable, MissingColumnError, DatabaseRetentionPolicy

lock = threading.Lock()


def run_locked(func):
    """decorator to run a function using a lock"""
    def wrapper(*args, **kwargs):
        with lock:
            func(*args, **kwargs)
    return wrapper


class TestDatabaseTable(TimeBasedDatabaseTable):
    @property
    def table_name(self) -> str:
        return 'TEST_DB_TABLE'

    @property
    def column_name_types(self) -> List[Tuple[str, str]]:
        return [
            ('datetime', 'DATE'),
            ('col_1', 'TEXT'),
            ('col_2', 'REAL'),
            ('col_3', 'BOOL'),
        ]

    def retrieve(self,
                 start_time: datetime = None,
                 end_time: datetime = None,
                 duration: timedelta = timedelta(hours=1),
                 ) -> List[List]:
        if duration is not None:
            if start_time is None and end_time is None:
                end_time = datetime.now()

            if start_time and end_time:
                pass
            elif start_time:
                end_time = start_time + duration
            elif end_time:
                start_time = end_time - duration
            command = self._select_query + " WHERE datetime BETWEEN :start_time AND :end_time"
            values = {
                'start_time': start_time,
                'end_time': end_time,
            }
        else:
            command = self._select_query
            values = {}
        with sqlite3.connect(self.file_path) as db:
            with closing(db.cursor()) as cur:
                cur.execute(command, values)
                rows = cur.fetchall()
        out = []
        for row in rows:
            # cast each value to the types they should be
            try:
                dt = datetime.strptime(row[0], "%Y-%m-%d %H:%M:%S.%f")
            except ValueError:
                dt = datetime.strptime(row[0], "%Y-%m-%d %H:%M:%S")
            col_1 = str(row[1])
            col_2 = int(row[2])
            col_3 = bool(row[3])
            # add the row of values to the output
            r = [dt, col_1, col_2, col_3]
            out.append(r)
        return out


class BadTestTimeBasedDatabaseTable(TimeBasedDatabaseTable):
    @property
    def column_name_types(self) -> List[Tuple[str, str]]:
        return [
            ('col_1', 'TEXT'),
            ('col_2', 'REAL'),
            ('col_3', 'BOOL'),
        ]

    @property
    def table_name(self) -> str:
        return 'A_BAD_TEMPORAL_TEST_DB_TABLE'

    def retrieve(self,
                 start_time: datetime = None,
                 end_time: datetime = None,
                 duration: timedelta = timedelta(hours=1),
                 ) -> List[List]:
        # just pass this because we dont expect to be able to instantiate a class instance
        pass


class AnotherTestDatabaseTable(DatabaseTable):
    @property
    def column_name_types(self) -> List[Tuple[str, str]]:
        return [
            ('col_1', 'TEXT'),
            ('col_2', 'REAL'),
            ('col_3', 'BOOL'),
        ]

    @property
    def table_name(self) -> str:
        return 'ANOTHER_TEST_DB_TABLE'


class TestDatabaseTable31DayRetentionPolicy(DatabaseRetentionPolicy):
    """A policy to only retain entries made within the last 31 days"""
    @property
    def _retention_policy(self) -> str:
        policy = "delete from TEST_DB_TABLE where datetime not between :start_time and :end_time"
        return policy

    @property
    def _retention_policy_values(self) -> Dict:
        end_time = datetime.now()
        start_time = end_time - timedelta(days=31)
        values = {
            'start_time': start_time,
            'end_time': end_time,
        }
        return values


class TestDBTools(unittest.TestCase):

    db_path: str = 'test_db.db'
    db_time_based_table: TimeBasedDatabaseTable = None
    db_retention_policy_31_days: TestDatabaseTable31DayRetentionPolicy = TestDatabaseTable31DayRetentionPolicy(enabled=False)

    def setUp(self) -> None:
        if Path(self.db_path).exists():
            # delete any existing db files
            Path(self.db_path).unlink()
        self.db_time_based_table = TestDatabaseTable(file_path=self.db_path)

    def test_db_extension(self):
        """tests db basic setting"""
        self.assertIsInstance(self.db_time_based_table.file_path, pathlib.Path, 'ensure table path is pathlib.Path')
        self.assertEqual(self.db_time_based_table.file_path.suffix, '.db', 'ensure extension exists')

    @run_locked
    def test_db_basic_properties(self):
        table_names = self.db_time_based_table.db_table_names
        self.assertEqual(1, len(table_names), 'there should only be 1 table in the database')
        self.assertEqual('TEST_DB_TABLE', table_names[0], 'the database table name should match')

        column_names = self.db_time_based_table.column_names
        self.assertEqual(4, len(column_names), 'there should only be 3 columns in the database')
        for index in range(4):
            if index == 0:
                self.assertEqual('datetime', column_names[index])
            else:
                col_number = index
                self.assertEqual(f'col_{col_number}', column_names[index])

    @run_locked
    def test_insert_retrieve_remove(self):
        wait_time = 1  # second, wait time between inserting data into the table

        # test inserting into the database table
        insert_1_time = datetime.now()
        insert_1 = (insert_1_time, 'insert 1', 1, True)
        self.db_time_based_table.insert(insert_1)
        expected_1 = [list(insert_1)]
        retrieve_1 = self.db_time_based_table.retrieve()
        self.assertEqual(expected_1, retrieve_1, 'ensure only 1 row with the correct values')
        time.sleep(wait_time)

        insert_2_time = datetime.now()
        insert_2 = (insert_2_time, 'insert 2', 2, True)
        self.db_time_based_table.insert(insert_2)
        expected_1_and_2 = [list(insert_1), list(insert_2)]
        retrieve_1_and_2 = self.db_time_based_table.retrieve()
        self.assertEqual(expected_1_and_2, retrieve_1_and_2, 'ensure both rows returned with the correct values')
        time.sleep(wait_time)

        insert_3_time = datetime.now()
        insert_3 = (insert_3_time, 'insert 3', 3, False)
        self.db_time_based_table.insert(insert_3)
        time.sleep(wait_time)

        insert_4_time = datetime.now()
        insert_4 = (insert_4_time, 'insert 4', 4, False)
        self.db_time_based_table.insert(insert_4)
        time.sleep(wait_time)

        insert_5_time = datetime.now()
        insert_5 = (insert_5_time, 'insert 5', 5, False)
        self.db_time_based_table.insert(insert_5)
        expected_1_and_2_and_3_and_4_and_5 = [list(insert_1), list(insert_2), list(insert_3), list(insert_4),
                                              list(insert_5)]
        retrieve_1_and_2_and_3_and_4_and_5 = self.db_time_based_table.retrieve()
        self.assertEqual(expected_1_and_2_and_3_and_4_and_5,
                         retrieve_1_and_2_and_3_and_4_and_5,
                         'ensure all five rows returned with the correct values')

        # test that retrieving data within specific timeframes work
        expected_1_and_2_and_3 = [list(insert_1), list(insert_2), list(insert_3)]
        retrieve_1_and_2_and_3 = self.db_time_based_table.retrieve(start_time=insert_1_time, end_time=insert_3_time)
        self.assertEqual(expected_1_and_2_and_3, retrieve_1_and_2_and_3,
                         'ensure retrieving from a start time & end time works')

        expected_2_and_3_and_4 = [list(insert_2), list(insert_3), list(insert_4)]
        # use duration of wait_time * 2.5 instead of 2 as a bit of a fudge factor for the time it takes to run code
        retrieve_2_and_3_and_4 = self.db_time_based_table.retrieve(start_time=insert_2_time, duration=timedelta(seconds=wait_time * 2.5))
        self.assertEqual(expected_2_and_3_and_4,
                         retrieve_2_and_3_and_4,
                         'ensure retrieving from a start time & duration works')

        expected_3_and_4 = [list(insert_3), list(insert_4)]
        # use duration of wait_time * 2.5 instead of 2 as a bit of a fudge factor for the time it takes to run code
        retrieve_3_and_4 = self.db_time_based_table.retrieve(end_time=insert_4_time, duration=timedelta(seconds=wait_time * 1.5))
        self.assertEqual(expected_3_and_4, retrieve_3_and_4,
                         'ensure retrieving from an end time & duration works')

        # test retaining rows only within specific times
        self.db_time_based_table.retain(start_time=insert_2_time, end_time=insert_5_time)
        expected_2_and_3_and_4_and_5 = [list(insert_2), list(insert_3), list(insert_4), list(insert_5)]
        retrieve_2_and_3_and_4_and_5 = self.db_time_based_table.retrieve()
        self.assertEqual(expected_2_and_3_and_4_and_5, retrieve_2_and_3_and_4_and_5,
                         'ensure retrieving from a start time & end time works')

        self.db_time_based_table.retain(start_time=insert_2_time, duration=timedelta(seconds=wait_time * 2.5))
        expected_2_and_3_and_4 = [list(insert_2), list(insert_3), list(insert_4)]
        retrieve_2_and_3_and_4 = self.db_time_based_table.retrieve()
        self.assertEqual(expected_2_and_3_and_4, retrieve_2_and_3_and_4,
                         'ensure retrieving from a start time & duration works')

        self.db_time_based_table.retain(end_time=insert_4_time, duration=timedelta(seconds=wait_time * 1.5))
        expected_3_and_4 = [list(insert_3), list(insert_4)]
        retrieve_3_and_4 = self.db_time_based_table.retrieve()
        self.assertEqual(expected_3_and_4, retrieve_3_and_4,
                         'ensure retrieving from an end time & duration works')

        # test retrieve and retain with default values works
        insert_31_days_ago_time = datetime.now() - timedelta(days=31, seconds=1)
        insert_31_days_ago = (insert_31_days_ago_time, 'insert 31 days ago', -31, True)
        self.db_time_based_table.insert(insert_31_days_ago)

        retrieve_3_and_4 = self.db_time_based_table.retrieve()
        self.assertEqual(expected_3_and_4, retrieve_3_and_4,
                         'ensure retrieving with defaults end time now duration 1 hour works')

        expected_3_and_4_and_neg31 = [list(insert_3), list(insert_4), list(insert_31_days_ago)]
        retrieve_3_and_4_and_neg31 = self.db_time_based_table.retrieve(duration=timedelta(days=32))
        self.assertEqual(expected_3_and_4_and_neg31, retrieve_3_and_4_and_neg31,
                         'ensure retrieving older than the default duration of 31 days works')

        self.db_time_based_table.retain()
        retrieve_3_and_4 = self.db_time_based_table.retrieve(duration=None)
        self.assertEqual(expected_3_and_4, retrieve_3_and_4,
                         'ensure retaining with default end time now duration 1 hour works')

    @run_locked
    def test_retention_policy(self):
        self.assertFalse(self.db_retention_policy_31_days.enabled, 'retention policy should not be enabled')

        # initially add values into the database table
        insert_1_time = datetime.now()
        insert_1 = (insert_1_time, 'insert 1', 1, True)
        self.db_time_based_table.insert(insert_1)

        insert_almost_31_days_ago_time = datetime.now() - timedelta(days=30, hours=23, minutes=59)
        insert_almost_31_days_ago = (insert_almost_31_days_ago_time, 'insert almost 31 days ago', -31, True)
        self.db_time_based_table.insert(insert_almost_31_days_ago)

        insert_32_days_ago_time = datetime.now() - timedelta(days=32)
        insert_32_days_ago = (insert_32_days_ago_time, 'insert 32 days ago', -32, True)
        self.db_time_based_table.insert(insert_32_days_ago)

        # run the policy on the database table, but dont enable the policy so the values should not be removed
        expected_1_and_almost_31_and_32 = [list(insert_1), list(insert_almost_31_days_ago), list(insert_32_days_ago)]
        self.assertFalse(self.db_retention_policy_31_days(self.db_time_based_table),
                         'policy should not run because it is not enabled')
        retrieved_1_and_almost_31_and_32 = self.db_time_based_table.retrieve(duration=None)
        self.assertEqual(expected_1_and_almost_31_and_32, retrieved_1_and_almost_31_and_32,
                         'all entries should be in the database if duration is None')

        # run the policy on the database table with the policy enabled so values older than 31 days should be removed
        self.db_retention_policy_31_days.enabled = True
        self.assertTrue(self.db_retention_policy_31_days.enabled, 'retention policy should now be enabled')
        expected_1_and_almost_31 = [list(insert_1), list(insert_almost_31_days_ago)]
        policy_ran = self.db_retention_policy_31_days(self.db_time_based_table)
        self.assertTrue(policy_ran,
                        'policy should run because it is enabled')
        retrieved_1_and_almost_31 = self.db_time_based_table.retrieve(duration=None)
        self.assertEqual(expected_1_and_almost_31, retrieved_1_and_almost_31,
                         'the 32 day old entry should be removed and only the other two entries should remain')

    @run_locked
    def test_adding_second_table(self):
        # test adding a second table to the database
        another_db_table = AnotherTestDatabaseTable(file_path=self.db_path)

        table_names = another_db_table.db_table_names
        self.assertEqual(2, len(table_names), 'there should only be 2 tables in the database')
        self.assertEqual(['TEST_DB_TABLE', 'ANOTHER_TEST_DB_TABLE'], table_names,
                         'the database table names should match')

    @run_locked
    def test_dont_add_a_bad_table(self):
        # should not be able to instantiate this TimeBasedDatabaseTable because datetime is not a column
        with self.assertRaises(MissingColumnError) as cm:
            BadTestTimeBasedDatabaseTable(file_path=self.db_path)
        self.assertEqual(cm.exception.msg, 'datetime must be a column in the table', 'ensure MissingColumnError was thrown')

    def test_basic_protection(self):
        """tests the basic sqlite injection prevention"""
        for attr in ['_base_query', 'table_name']:
            self.assertRaises(
                AttributeError,
                setattr,
                self.db_time_based_table,
                attr,
                'malicious value'
            )





import logging

from typing import Union, Callable, Iterable

from .basic import Action
from .configured import ConfiguredAction, ConfiguredActionList, TrackedActionList, TrackedAction
from .operators import decision


logger = logging.getLogger(__name__)


class ConfiguredDecision(ConfiguredAction):
    UUID_PREFIX = 'cda-'

    def __init__(self,
                 *logic_checks: Union[Callable, Action, str, ConfiguredAction],
                 configuration_name: str = None,
                 **kwargs,
                 ):
        """
        Class for configuring a logic tree. The logic checks will iterate until an action return is True. When that
        condition is met, the next action of that logic check will be returned from the executor.

        :param logic_checks: logic check actions. These must return a bool
        :param configuration_name: name for the configuration
        """
        ConfiguredAction.__init__(
            self,
            action=decision,
            configuration_name=configuration_name,
            **kwargs,
        )
        self._logic_sequence: ConfiguredActionList = ConfiguredActionList()
        self.logic_sequence = logic_checks

    @property
    def logic_sequence(self) -> ConfiguredActionList:
        """list of logic check actions to be performed"""
        return self._logic_sequence

    @logic_sequence.setter
    def logic_sequence(self, value: Iterable):
        self._logic_sequence = ConfiguredActionList(*value)
        for action in self._logic_sequence:
            if action.action.returns_bool is False:
                logger.warning(
                    f'the action {action} does not type hint bool, using this action for decisions may result in '
                    f'unexpected errors'
                )

    def get_tracked_from_config(self, **kwargs) -> "TrackedDecision":
        """
        Generates a tracked instance from the configured action

        :param kwargs: additional kwargs that differ from the instance configuration
        :return: TrackedDecision instance.
        """
        return TrackedDecision(configuration=self, **kwargs)

    def as_dict(self) -> dict:
        dct = super(ConfiguredDecision, self).as_dict()
        dct['logic_sequence'] = self.logic_sequence.as_list()
        return dct


class TrackedDecision(TrackedAction):
    UUID_PREFIX = 'da-'

    def __init__(self,
                 configuration: Union[str, ConfiguredDecision],
                 **kwargs,
                 ):
        """
        When called, given multiple ConfiguredActions, each of which their action must return a bool, sequentially run
        each ConfiguredAction, and for the first ConfiguredAction that returns True, return the next_action of the
        ConfiguredAction

        :param args: ConfiguredActions that have actions that return a bool
        :param configuration_name: optional name for the configuration (for key retrieval)
        """
        if isinstance(configuration, ConfiguredDecision) is False:
            configuration = ConfiguredDecision.class_instance_by_id(configuration)
        self._configuration: ConfiguredDecision = None

        TrackedAction.__init__(
            self,
            configuration=configuration,
            **kwargs,
        )
        self._next_action: ConfiguredAction = None

    def __call__(self, *args, **kwargs):
        logger.info('beginning execution of logic tree')
        # convert configured actions to tracked actions
        actions = TrackedActionList(*self.configuration.logic_sequence)
        super(TrackedDecision, self).__call__(
            *actions,
            **kwargs,
        )
        logger.info(f'logic execution complete')

    @property
    def configuration(self) -> ConfiguredDecision:
        """the configured sequence associated with the instance"""
        return self._configuration

    @configuration.setter
    def configuration(self, value: ConfiguredDecision):
        if isinstance(value, ConfiguredDecision) is False:
            raise TypeError(f'a ConfiguredDecision must be used as a configuration for {self.__class__.__name__}')
        self._configuration = value
        self.update_args(*self._configuration.args)
        self.update_kwargs(*self._configuration.kwargs)

    def copy_configuration(self) -> "TrackedDecision":  # todo update docstring
        """
        Copies the configuration of a ExecutionSequence instance.

        :return: ExecutionSequence instance with identical configuration
        """
        return self.__class__(
            self.configuration,
            **self.kwargs,
        )

    @property
    def next_action(self) -> ConfiguredAction:
        """next action as determined by the logic tree"""
        return self._action_return

    def as_dict(self) -> dict:
        dct = super(TrackedDecision, self).as_dict()
        if self.triggered is True:
            dct['logic_checks'] = dct['arguments']
            del dct['arguments']
        return dct



"""
Tools for managing sequences of actions (execution) or linked chains of actions (automation).

For linear sequences of actions, it is recommended to use the `.sequencing.execution` module. For linked actions
which are linked by `.next_action` attributes use the `.sequencing.automation` module.
"""
from .execution import ExecutionSequence, ConfiguredSequence
from .automation import Automation, ConfiguredAutomation

__all__ = [
    'ConfiguredSequence',
    'ExecutionSequence',
    'ConfiguredAutomation',
    'Automation',
]



import threading
import time
import unittest
import operator
from hein_control.states import (ComponentState, OperationalState, ensure_component_online, ensure_component_inactive,
                                 BadComponentState, GenericActionState, TemporaryOperationalCallback,
                                 TemporaryActionCallback)
from hein_control.states.abc import State, StateSet
from hein_control.action.configured import ConfiguredAction


class CustomOperationalState(StateSet):
    ERROR = State(-1)
    OFFLINE = State(0)
    ONLINE = State(1)
    AMAZING = State(2)


class CustomActiveState(StateSet):
    IDLE = State(0)
    ACTIVE = State(1)
    HYPERACTIVE = State(2)
    JUMP_JUMP = 3
    BOOGIE_WOOGIE = State(4)
    # can you tell I have a two year old?


class TestComponent(ComponentState):
    action_states = CustomActiveState
    operational_states = CustomOperationalState

    @ensure_component_online
    def only_run_online(self):
        return

    @ensure_component_inactive
    def only_run_inactive(self):
        return


class TestComponentStates(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.lock = threading.Lock()
        cls.component_one = TestComponent('test_component1', 1)
        cls.component_two = ComponentState('test_component2', 1)
        cls.component_three = ComponentState('test_component3', 1)

    def set_and_check_state(self, component: ComponentState, attribute: str, state: StateSet):
        """generic tester for action and operational state"""
        # test by value
        setattr(component, attribute, state.value)
        self.assertIs(getattr(component, attribute), state)
        # test by name
        setattr(component, attribute, state.name)
        self.assertIs(getattr(component, attribute), state)
        # test name capitalization support
        setattr(component, attribute, state.name.lower())
        self.assertIs(getattr(component, attribute), state)
        # test direct state setting
        setattr(component, attribute, state)
        self.assertIs(getattr(component, attribute), state)

    def set_and_check_operational_state(self, component: ComponentState, state: OperationalState):
        """tests the setting of an operational state by state, value, and name"""
        # test property setting
        self.set_and_check_state(component, 'current_operational_state', state)
        # test method updating
        component.update_operational_state(state)
        self.assertIs(component.current_operational_state, state)

    def set_and_check_action_state(self, component: ComponentState, state: OperationalState):
        """tests the setting of an active state by state, value, and name"""
        # test property setting
        self.set_and_check_state(component, 'current_action_state', state)
        # test method updating
        component.update_action_state(state)
        self.assertIs(component.current_action_state, state)

    def test_component_state_setting(self):
        """tests setting and retrieval of component states"""
        with self.lock:
            for state in CustomOperationalState:
                self.set_and_check_operational_state(self.component_one, state)

    def test_online_only(self):
        """tests the online check decorator functionality"""
        with self.lock:
            try:
                # check that the method runs
                self.component_one.only_run_online()
                # set to offline state
                self.component_one.current_operational_state = 0
                self.assertRaises(BadComponentState, self.component_one.only_run_online)
                # set to error state
                self.component_one.current_operational_state = -1
                self.assertRaises(BadComponentState, self.component_one.only_run_online)
            finally:
                self.component_one.current_operational_state = OperationalState.ONLINE

    def test_component_activity_setting(self):
        """tests the setting of the component activity"""
        with self.lock:
            for state in GenericActionState:
                self.set_and_check_action_state(self.component_two, state)

    def test_custom_activity_setting(self):
        """tests components with custom activity support"""
        with self.lock:
            for state in CustomActiveState:
                self.set_and_check_action_state(self.component_one, state)

    def test_inactive_only(self):
        """tests the inactive-only decorator functionality"""
        with self.lock:
            try:
                # set to active state
                for state in CustomActiveState:
                    self.component_one.current_action_state = state
                    if state == 0:
                        # check that the method runs
                        self.component_one.only_run_inactive()
                    else:
                        self.assertRaises(BadComponentState, self.component_one.only_run_inactive)
            finally:
                self.component_one.current_action_state = 0

    def update_component_state(self, state):
        """sets component one to an error state after 1 second"""
        time.sleep(1)
        self.component_one.current_operational_state = state

    def get_component_state_thread(self, state: int = -1):
        """retrieves a thread which sets the first component to the specified state"""
        return threading.Thread(
            target=self.update_component_state,
            daemon=True,
            args=[state],
        )

    def update_component_action_state(self, state):
        """sets the component action to the defined state after 1 second"""
        time.sleep(1.)
        self.component_one.current_action_state = state

    def get_component_action_state_thread(self, state: int = -1):
        """retrieves a thread which sets the first component to the specified state"""
        return threading.Thread(
            target=self.update_component_action_state,
            daemon=True,
            args=[state],
        )

    def test_self_waiting(self):
        """tests waiting on a component where the error flag is switched mid-wait"""
        with self.lock:
            try:
                # ensure state monitor actually sleeps
                self.component_one.components_state_monitor_sleep(0.1)
                # tests that error raises
                first_thread = self.get_component_state_thread()
                first_thread.start()
                self.assertRaises(
                    BadComponentState,
                    self.component_one.components_state_monitor_sleep,
                    2
                )

            finally:
                self.component_one.current_operational_state = 1

    def test_adjacent_waiting(self):
        """tests a wait where the error flag is flipped on another component"""
        with self.lock:
            try:
                # ensure state monitor actually sleeps
                self.component_two.components_state_monitor_sleep(0.1)
                # perform wait on another component and ensure that bad state propagates to an error
                second_thread = self.get_component_state_thread()
                second_thread.start()
                self.assertRaises(
                    BadComponentState,
                    self.component_two.components_state_monitor_sleep,
                    2
                )

            finally:
                self.component_one.current_operational_state = 1

    def test_solo_monitor(self):
        """tests a wait where a single component is monitored for an error state (ignores errors from other components"""
        with self.lock:
            try:
                # perform wait on an third component, specifically only monitoring that component
                #   component 1 is still in an error state for this test
                third_thread = self.get_component_state_thread()
                third_thread.start()
                self.component_three.components_state_monitor_sleep(
                    1.5,
                    'test_component3'
                )
            finally:
                self.component_one.current_operational_state = 1

    def test_wait_for_operational(self):
        """tests the functionality for waiting for a component to be operational"""
        with self.lock:
            try:
                # set to offline
                self.component_one.current_operational_state = 0
                thread = self.get_component_state_thread(1)
                thread.start()
                self.component_one.wait_for_component_operational_state(1)

                # test timeout
                self.component_one.current_operational_state = 0
                self.assertRaises(
                    TimeoutError,
                    self.component_one.wait_for_component_operational_state,
                    1,
                    timeout=0.5,
                )

            finally:
                self.component_one.current_operational_state = 1

    def test_wait_for_action(self):
        """tests the functionality for waiting for a component to be ___"""
        with self.lock:
            try:
                self.component_one.current_action_state = 0
                thread = self.get_component_action_state_thread(1)
                thread.start()
                self.component_one.wait_for_component_action_state(1)

                # test timeout
                self.component_one.current_action_state = 0
                self.assertRaises(
                    TimeoutError,
                    self.component_one.wait_for_component_action_state,
                    1,
                    timeout=0.5,
                )

            finally:
                self.component_one.current_action_state = 0

    def test_timeout(self):
        """tests the timeout functionality for the flag check wait"""
        def always_false():
            """Always blue! Always blue! Always blue! Always blue! Always blue!"""
            return False
        with self.lock:
            self.assertRaises(
                TimeoutError,
                self.component_one.state_monitor_flag_check,
                always_false,
                timeout=1.
            )
            # ensure that the function runs without a timeout specified
            self.component_one.state_monitor_flag_check(
                always_false,
                target_flag_state=False,
            )

    def test_action_state_retrieval(self):
        """tests internal aciton state retrieval"""
        for action_state in self.component_one.action_states:
            action_state: StateSet
            for key in [
                action_state,
                action_state.value,
                action_state.name,
                action_state.value.value if isinstance(action_state.value, State) else action_state.value,
            ]:
                self.assertIs(
                    self.component_one.get_action_state_from_key(key),
                    action_state,
                    f'check retrieval of {key}'
                )

    def test_operational_state_retrieval(self):
        """tests internal aciton state retrieval"""
        for action_state in self.component_one.operational_states:
            action_state: StateSet
            for key in [
                action_state,
                action_state.value,
                action_state.name,
                action_state.value.value if isinstance(action_state.value, State) else action_state.value,
            ]:
                self.assertIs(
                    self.component_one.get_operational_state_from_key(key),
                    action_state,
                    f'check retrieval of {key}'
                )

    def test_action_comparison_operator_override(self):
        """tests the comparison operator override capability for the action wait"""
        with self.lock:
            # test eq logic (simulates waiting for an action state to be in a specific place)
            comp_operator = operator.eq
            self.component_one.current_action_state = 0
            thread = self.get_component_action_state_thread(2)
            thread.start()
            self.assertRaises(
                TimeoutError,
                self.component_one.wait_for_component_action_state,
                desired_state=1,
                timeout=0.5,
                comparison_operator=comp_operator,

            )
            self.component_one.current_action_state = 0
            thread = self.get_component_action_state_thread(2)
            thread.start()
            self.component_one.wait_for_component_action_state(2, timeout=1.5, comparison_operator=comp_operator)

            # test lt logic (simulates waiting for an action state to indicate idle)
            comp_operator = operator.le
            self.assertRaises(
                TimeoutError,
                self.component_one.wait_for_component_action_state,
                0,
                timeout=0.5,
                comparison_operator=comp_operator,
            )
            thread = self.get_component_action_state_thread(0)
            thread.start()
            self.component_one.wait_for_component_action_state(
                0,
                timeout=1.5,
                comparison_operator=comp_operator,
            )

    def test_operational_comparison_operator_override(self):
        """tests the comparison operator override capability for the operational state wait"""
        with self.lock:
            # test eq logic (simulates waiting for an action state to be in a specific place)
            comp_operator = operator.eq
            self.component_one.current_operational_state = 0
            thread = self.get_component_state_thread(2)
            thread.start()
            self.assertRaises(
                TimeoutError,
                self.component_one.wait_for_component_operational_state,
                desired_state=1,
                timeout=0.5,
                comparison_operator=comp_operator,

            )
            self.component_one.current_operational_state = 0
            thread = self.get_component_state_thread(2)
            thread.start()
            self.component_one.wait_for_component_operational_state(2, timeout=1.5, comparison_operator=comp_operator)

            # test lt logic (simulates waiting for an action state to indicate idle)
            comp_operator = operator.le
            self.assertRaises(
                TimeoutError,
                self.component_one.wait_for_component_operational_state,
                0,
                timeout=0.5,
                comparison_operator=comp_operator,
            )
            thread = self.get_component_state_thread(0)
            thread.start()
            self.component_one.wait_for_component_operational_state(
                0,
                timeout=1.5,
                comparison_operator=comp_operator,
            )
            self.component_one.current_operational_state = 1


class TestStates(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.test_state_one = State(1)

    def test_meaning(self):
        """tests meaning setting of the State class"""
        string = 'test meaning string (docstring) for the state'
        state = State(1, meaning=string)
        self.assertTrue(hasattr(state, 'meaning'), 'ensure that a meaning was assigned')
        self.assertEqual(state.meaning, string, 'check that the state was correctly assigned')

    def test_comparisons(self):
        """tests integer state comparisons"""
        self.assertTrue(1 == self.test_state_one, 'check eq')
        self.assertTrue(2 > self.test_state_one, 'check gt')
        self.assertTrue(1 >= self.test_state_one, 'check ge eq')
        self.assertTrue(2 >= self.test_state_one, 'check ge gt')
        self.assertTrue(0 < self.test_state_one, 'check lt')
        self.assertTrue(1 <= self.test_state_one, 'check le eq')
        self.assertTrue(0 <= self.test_state_one, 'check le lt')

    def test_state_comparisons(self):
        """tests comparisons between State classes"""
        self.assertTrue(State(1) == self.test_state_one, 'check eq')
        self.assertTrue(State(2) > self.test_state_one, 'check gt')
        self.assertTrue(State(1) >= self.test_state_one, 'check ge eq')
        self.assertTrue(State(2) >= self.test_state_one, 'check ge gt')
        self.assertTrue(State(0) < self.test_state_one, 'check lt')
        self.assertTrue(State(1) <= self.test_state_one, 'check le eq')
        self.assertTrue(State(0) <= self.test_state_one, 'check le lt')

    def test_callbacks(self):
        """tests state callback de/registration and execution"""
        def banana():
            return 1234

        def melon():
            return 'asdf'

        # test registration
        self.test_state_one.register_callbacks(banana)
        self.test_state_one.register_post_state_callbacks(melon)
        self.assertEqual(len(self.test_state_one.pre_state_callbacks), 1, "ensure callback was added")
        self.assertEqual(len(self.test_state_one.post_state_callbacks), 1, 'ensure post callback was added')
        self.assertRaises(TypeError, self.test_state_one.register_callbacks, 1, 'check that non-callables are rejected')

        # test execution
        values = self.test_state_one.execute_pre_state_callbacks()
        self.assertEqual(len(values), 1, 'check that two returns were captured')
        self.assertEqual(values[0].action_return, 1234, 'check that first function value is correct')
        post_values = self.test_state_one.execute_post_state_callbacks()
        self.assertEqual(post_values[0].action_return, 'asdf', 'check that second function value is correct')

        # test deregistration
        self.test_state_one.deregister_callbacks(banana)
        self.assertEqual(len(self.test_state_one.pre_state_callbacks), 0, 'ensure callback was removed')
        self.assertFalse(self.test_state_one.has_pre_state_change_callbacks, 'ensure pre-state-change callback was removed')
        self.assertTrue(self.test_state_one.has_post_state_change_callbacks, 'ensure post-state-change callback was unaffected')
        self.assertFalse(banana in self.test_state_one.pre_state_callbacks, 'check that method was deregistered')

        self.test_state_one.deregister_callbacks(melon)
        self.assertFalse(self.test_state_one.has_callbacks)
        self.assertFalse(self.test_state_one.has_post_state_change_callbacks)
        self.assertFalse(self.test_state_one.has_pre_state_change_callbacks)


class FancyOperationalStates(StateSet):
    CRAZY_BAD = State(-9000)
    BAD = State(-1)
    OFFLINE = 0
    GOOD = State(1)
    GREAT = State(2)


class FancyTestComponent(ComponentState):
    operational_states = FancyOperationalStates

    bad_string = 'omg this super bad thing happened'
    good_string = 'never mind, things are fine'
    no_change = 'no change needed'
    allowable_jump = 'sure, you can do that'

    def __init__(self):
        """a test component which has """
        super(FancyTestComponent, self).__init__('fancy_test_component')
        self.real_bad_thing_happened = False
        self.operational_states.BAD.register_callbacks(self.critical_action)
        self.operational_states.GOOD.register_callbacks(self.restore_functionality)
        self.operational_states.CRAZY_BAD.register_callbacks(self.check_operational_state_jump)
        self.action_states.ACTIVE.register_callbacks(self.check_action_state_jump)

    def critical_action(self):
        """registers that a critical thing has happened"""
        self.real_bad_thing_happened = True
        return self.bad_string

    def restore_functionality(self):
        """restores functionality after a critical thing has happened"""
        if self.current_operational_state > 0:
            return self.no_change
        self.real_bad_thing_happened = False
        return self.good_string

    def check_operational_state_jump(self):
        """will raise an error if the state was already changed"""
        if self.current_operational_state == -9000:
            raise ValueError('this error will occur only if the state is changed before the callbacks are executed')
        return self.allowable_jump

    def check_action_state_jump(self):
        """will raise an error if the state was already changed"""
        if self.current_action_state == 1:
            raise ValueError('this error will occur only if the state is changed before the callbacks are executed')
        return self.allowable_jump


class TestComponentCallbacks(unittest.TestCase):
    """Tests automatic callback execution component state switching"""
    @classmethod
    def setUpClass(cls) -> None:
        cls.component = FancyTestComponent()
        cls.lock = threading.Lock()

    def test_registration(self):
        """ensure actions were registered as expected"""
        self.assertIsNotNone(
            self.component.operational_states.BAD.value.pre_state_callbacks.get_ca_by_action(
                self.component.critical_action
            )
        )
        self.assertTrue(
            self.component.operational_states.GOOD.value.pre_state_callbacks.get_ca_by_action(
                self.component.restore_functionality
            )
        )

    def test_callback_functionality(self):
        """tests callback functionality"""
        with self.lock:
            try:
                # change the operational state to a bad state; ensure callback was executed
                value = self.component.update_operational_state(-1)
                self.assertEqual(value[0].action_return, self.component.bad_string, 'ensure return value is captured')
                self.assertTrue(self.component.real_bad_thing_happened, 'ensure callback executed as expected')

                # change the operational state to a good one; ensure callback was executed
                value = self.component.update_operational_state(1)
                self.assertEqual(value[0].action_return, self.component.good_string, 'ensure return value is captured')
                self.assertFalse(self.component.real_bad_thing_happened, 'ensure callback executed as expected')

                # test condition check within restor function
                value = self.component.update_operational_state(1)
                self.assertEqual(value[0].action_return, self.component.no_change, 'ensure conditional checks work')

                # these will fail if the state is updated before the callbacks are executed
                value = self.component.update_operational_state(-9000)
                self.assertEqual(value[0].action_return, self.component.allowable_jump)
                value = self.component.update_action_state(1)
                self.assertEqual(value[0].action_return, self.component.allowable_jump)
            finally:
                self.component.update_operational_state(1)
                self.component.update_action_state(0)

    def pre_state_check_callback(self):
        """special callback which should be executed before a component state is changed"""
        self.assertEqual(self.component.current_operational_state, 1)

    def post_state_check_callback(self):
        """special callback which should be executed after a component state is changed"""
        self.assertEqual(self.component.current_operational_state, 2)

    def test_operational_callback_ordering(self):
        """
        tests that pre- and post- operational state change callbacks are executed in the correct place
        """
        with self.lock:
            try:
                self.component.update_operational_state(1)  # start from 1
                state = self.component.operational_states(2)
                state.register_pre_state_callbacks(self.pre_state_check_callback)
                state.register_post_state_callbacks(self.post_state_check_callback)
                self.component.update_operational_state(2)
                # deregister and check
                state.deregister_callbacks(self.pre_state_check_callback, self.post_state_check_callback)
                self.assertFalse(self.pre_state_check_callback in state.value.pre_state_callbacks)
                self.assertFalse(self.post_state_check_callback in state.value.post_state_callbacks)
            finally:
                self.component.update_operational_state(1)
                self.component.update_action_state(0)

    def pre_action_state_check_callback(self):
        """special callback which should be executed before a component state is changed"""
        self.assertEqual(self.component.current_action_state, 0)

    def post_action_state_check_callback(self):
        """special callback which should be executed after a component state is changed"""
        self.assertEqual(self.component.current_action_state, 1)

    def test_alias_setting(self):
        """tests alias setting for a state"""
        state = State(1, alias='banana')
        self.assertEqual(state.alias, 'banana', 'test alias was set from instantiation')
        state.alias = 'melon'
        self.assertEqual(state.alias, 'melon', 'test alias updating')

        class TestStateSet(StateSet):
            MELON = State(1)
            BANANA = 2

        test_state = TestStateSet(1)
        self.assertEqual(test_state.name, 'MELON', 'test base name')
        self.assertEqual(test_state.get_name(), 'MELON', 'test default')
        self.assertEqual(TestStateSet(2).get_name(), 'BANANA', 'test name retrieval from integer')
        test_state.set_alias('delicious')
        self.assertEqual(test_state.value.alias, 'delicious', 'test alias setting passthrough')
        self.assertEqual(test_state.get_name(), 'delicious', 'test name retrieval is updated')

    def test_action_callback_ordering(self):
        """
        tests that pre- and post- operational state change callbacks are executed in the correct place
        """
        with self.lock:
            try:
                self.component.update_action_state(0)
                state = self.component.action_states(1)
                state.register_pre_state_callbacks(self.pre_action_state_check_callback)
                state.register_post_state_callbacks(self.post_action_state_check_callback)
                self.component.update_action_state(1)
                # deregister and check
                state.deregister_callbacks(self.pre_action_state_check_callback, self.post_action_state_check_callback)
                self.assertFalse(self.pre_action_state_check_callback in state.value.pre_state_callbacks)
                self.assertFalse(self.post_action_state_check_callback in state.value.post_state_callbacks)
            finally:
                self.component.update_operational_state(1)
                self.component.update_action_state(0)

    def test_callback_deletion(self):
        """tests deletion functionality of callbacks"""
        def some_one_off_function():
            pass
        with self.lock:
            self.component.operational_states(1).value.register_callbacks(some_one_off_function)
            ca = self.component.operational_states(1).value.pre_state_callbacks[-1]
            self.assertIsNotNone(ConfiguredAction.class_instance_by_id(ca.name), 'check that retrieval works')
            self.component.operational_states(1).value.deregister_callbacks(
                some_one_off_function,
                remove_configurations=True,
            )
            self.assertNotIn(
                ca, self.component.operational_states(1).value.pre_state_callbacks,
                'ensure method was removed from callback list'
            )
            self.assertRaises(  # ensure that the configured action was removed as expected
                NameError,
                ConfiguredAction.class_instance_by_id,
                ca.name,
            )


class SoManyStates(StateSet):
    SUPER_BAD = State(-10)
    QUITE_BAD = State(-2)
    BAD = State(-1)
    NOT_GREAT = State(0)
    OK = State(1)
    GOOD = State(2)
    GREAT = State(3)
    EXCELLENT = State(10)


class AllTheStates(ComponentState):
    operational_states = SoManyStates
    action_states = CustomActiveState


class TestTemporaryCallback(unittest.TestCase):
    change_state_wait_return = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.component = AllTheStates('so many states')
        cls.lock = threading.Lock()

    @staticmethod
    def custom_callback():
        raise ValueError('oh no the worst thing has happened!')

    def check_assigned(self, fn, threshold=0):
        """performs the appropriate assignment checks depending on the threshold state"""
        for state in self.component.operational_states:
            if state <= threshold:
                self.assertTrue(state.has_callbacks, f'ensure a callback is assigned to {state}')
                self.assertTrue(
                    state.value.pre_state_callbacks.action_registered_in_list(fn),
                    f'ensure correct callback was assigned to {state}'
                )
            else:
                self.assertFalse(state.has_callbacks, 'ensure callback was not assigned to states above threshold')
                self.assertFalse(
                    state.value.pre_state_callbacks.action_registered_in_list(fn),
                    f'ensure callback was not assigned to {state}'
                )

    def check_unassigned(self, fn):
        """performs a check to ensure that the provided function has been unassigned from all component operational states"""
        for state in self.component.operational_states:
            self.assertFalse(state.has_callbacks, f'ensure no callbacks are registered to {state}')

    def test_default_callback(self):
        """tests setting of the default callback method"""
        with self.lock:
            try:
                with TemporaryOperationalCallback(self.component) as tcb:
                    self.assertIsInstance(tcb, TemporaryOperationalCallback, 'ensure callback TemporaryOperationalCallback instance is returned')
                    self.check_assigned(tcb.callback_capture)
                    self.component.update_operational_state(-1)
                    self.assertTrue(tcb.callback_encountered, 'ensure callback was encountered')
                for state in self.component.operational_states:
                    self.assertFalse(
                        tcb.callback_capture in state.value.pre_state_callbacks,
                        'ensure callback was removed'
                    )
                    self.assertFalse(state.has_callbacks)
            finally:
                self.component.update_operational_state(1)

    def test_threshold_setting(self):
        """tests that the callback was only applied to states at or below the threshold"""
        with self.lock:
            for threshold_state in self.component.operational_states:
                with TemporaryOperationalCallback(self.component, threshold_state=threshold_state) as tcb:
                    self.check_assigned(tcb.callback_capture, threshold_state)
                self.check_unassigned(tcb.callback_capture)

    def test_custom_callback(self):
        """tests custom callback setting"""
        with self.lock:
            try:
                with TemporaryOperationalCallback(self.component, callbacks=self.custom_callback) as tcb:
                    self.check_assigned(self.custom_callback)
                    self.change_state_wait()
                    self.assertIsNotNone(self.change_state_wait_return)
                    # load specific tracked actions
                    callback_capture, custom_callback = self.change_state_wait_return
                    self.assertEqual(callback_capture.status_code, 3, 'ensure callback capture was triggered')
                    self.assertEqual(custom_callback.status_code, -1, 'ensure error was raised by custom callback')
                    self.assertIsNotNone(custom_callback.error_details, 'ensure error details were set')
                    self.assertIsInstance(custom_callback.error_details, ValueError, 'ensure details are error instance')

                self.assertTrue(tcb.callback_encountered)
                # ensure callback was unassigned after exiting
                for state in self.component.operational_states:
                    self.assertFalse(state.has_callbacks)
                    self.assertFalse(tcb.callback_capture in state.value.pre_state_callbacks)
            finally:
                self.component.update_operational_state(1)

    def change_state_wait(self):
        """
        waits for some time and changes the component operational state. This should trigger an error if it is executed
        inside a TCB context
        """
        time.sleep(0.1)
        self.change_state_wait_return = self.component.update_operational_state(-1)

    def check_registered(self, lst, fn):
        """checks that the provided callable was registered in the provided ConfiguredActionList"""
        self.assertTrue(
            lst.action_registered_in_list(fn),
            f'ensure {fn} is registered in {lst}'
        )

    def check_not_registered(self, lst, fn):
        """checks that the provided callable is not registered in the provided ConfiguredActionList"""
        self.assertFalse(
            lst.action_registered_in_list(fn),
            f'ensure {fn} is not registered in {lst}'
        )

    def test_action_callback(self):
        """tests action callback setting"""
        threshold_state = self.component.action_states(2)
        with TemporaryActionCallback(self.component, callbacks=[self.custom_callback], threshold_state=threshold_state) as tcb:
            self.assertIsInstance(tcb, TemporaryActionCallback, 'ensure TemporaryActionCallback is returned')
            for state in self.component.action_states:
                if state <= threshold_state:
                    self.assertTrue(state.has_callbacks, f'ensure callback was assigned to {state}')
                    self.check_registered(state.value.pre_state_callbacks, self.custom_callback)
                elif hasattr(state.value, 'callbacks'):
                    self.assertFalse(state.has_callbacks, 'ensure callback was not assigned to states above threshold')
                    self.check_not_registered(state.value.pre_state_callbacks, self.custom_callback)

    def test_alias_retrieval(self):
        """tests retrieval of a ComponentState by a state alias"""
        state = self.component.operational_states.NOT_GREAT
        alias = 'so_so'
        state.set_alias(alias)
        self.assertIs(state, self.component.operational_states(alias), 'test retrieval of StateSet by alias')
        self.assertIs(state, self.component.operational_states(state.name), 'ensure retrieval by formal name still works')

    def test_action_removal(self):
        """tests action removal of temporary callbacks"""
        def some_other_action():
            pass

        with TemporaryOperationalCallback(self.component,
                                          callbacks=[some_other_action],
                                          remove_configurations_on_exit=True
                                          ) as tcb:
            capture_action_name = self.component.operational_states(0).value.pre_state_callbacks[0].name
            other_action_name = self.component.operational_states(0).value.pre_state_callbacks[1].name
            self.assertIsNotNone(ConfiguredAction.class_instance_by_id(capture_action_name))
            self.assertIsNotNone(ConfiguredAction.class_instance_by_id(other_action_name))

        # ensure that actions have been fully deregistered
        self.assertRaises(
            NameError,
            ConfiguredAction.class_instance_by_id,
            capture_action_name
        )
        self.assertRaises(
            NameError,
            ConfiguredAction.class_instance_by_id,
            other_action_name,
        )



import warnings

import time
import logging
import threading
import operator

from typing import Union, MutableMapping, List, Callable, Type, Optional, Iterable
from functools import wraps
from abc import ABC

from ..mixin.reg import InstanceRegistry


from .abc import state_logger, OperationalState, GenericActionState, State, StateSet


class BadComponentState(Exception):
    def __init__(self, *components: 'ComponentState', threshold_state: StateSet):
        """
        Exception signifying that bad state was encountered on the provided components

        :param components: components which have a bad state
        :param threshold_state: threshold state
        """
        self.components = components
        super(BadComponentState, self).__init__(
            f'a bad state was encountered with the component{"s" if len(components) > 0 else ""}: ' +
            ', '.join([
                f'{component.component_name}: {component.current_operational_state.name}'
                for component in components
            ])
            + f'; threshold state: {threshold_state.name}'
        )


def ensure_component_online(fn: Callable):
    """class instance decorator which will check for an online state before executing the decorated function"""
    @wraps(fn)
    def decorated(self: 'ComponentState', *args, **kwargs):
        if self.current_operational_state <= 0:
            state_logger.error(
                f'the component {self.component_name} is in operational state {self.current_operational_state.name} '
                f'(required: {OperationalState.ONLINE}) '
                f'when calling {fn.__name__}'
            )
            raise BadComponentState(self, threshold_state=OperationalState.OFFLINE)
        return fn(self, *args, **kwargs)

    return decorated


def ensure_component_inactive(fn: Callable):
    """
    class instance decorator which will check that the component is inactive before
    executing the decorated function
    """
    @wraps(fn)
    def decorated(self: 'ComponentState', *args, **kwargs):
        if self.current_action_state != 0:
            # retrieve the expected idle state enumeration
            expected_idle_state = self.current_action_state.__class__(0)
            state_logger.error(
                f'the component {self.component_name} is in action state {self.current_action_state.name} '
                f'(required: {expected_idle_state.name}) when calling {fn.__name__}'
            )
            raise BadComponentState(self, threshold_state=expected_idle_state)
        return fn(self, *args, **kwargs)

    return decorated


class TemporaryCallbackSetter(ABC):
    # state set to temporarily assign callbacks to
    _state_set_attribute: str = None

    def __init__(self,
                 *components: 'ComponentState',
                 callbacks: Union[Callable, Iterable[Callable]] = None,
                 threshold_state: StateSet = OperationalState.OFFLINE,
                 pre_state_change: bool = True,
                 remove_configurations_on_exit: bool = False,
                 ):
        """
        Context manager for temporarily applying callbacks to the provided components

        :param components: components to monitor
        :param callbacks: custom callback(s) to apply
        :param threshold_state: threshold state to apply callbacks at and below
        :param pre_state_change: whether the callbacks should be added to pre- or post-state change callback lists
        :param remove_configurations_on_exit: option to remove configurations on exit
            (useful for one-off registered methods that will never be used again)
        """
        self.components = components
        self.threshold_state = threshold_state
        self.pre_state_change = pre_state_change
        self.remove_on_exit = remove_configurations_on_exit
        if callbacks is None:
            callbacks = []
        elif callable(callbacks):  # cast if it's only a single callback
            callbacks = [callbacks]
        self.callbacks: Iterable[Callable] = callbacks
        # internal flag in the event that the callback is triggered
        self._callback_encountered = threading.Event()

    def __enter__(self) -> 'TemporaryCallbackSetter':
        for component in self.components:
            for state in getattr(component, self._state_set_attribute):
                if state <= self.threshold_state:
                    state.register_callbacks(self.callback_capture, *self.callbacks, pre_state=self.pre_state_change)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for component in self.components:
            for state in getattr(component, self._state_set_attribute):
                if state <= self.threshold_state:
                    state.deregister_callbacks(self.callback_capture, remove_configurations=True)
                    state.deregister_callbacks(
                        *self.callbacks,
                        remove_configurations=self.remove_on_exit,
                    )

    @property
    def callback_encountered(self) -> bool:
        """whether a callback has been encountered"""
        return self._callback_encountered.is_set()

    def callback_capture(self):
        """a default callback which notes that the callback was called (useful for flag checking)"""
        self._callback_encountered.set()


class TemporaryOperationalCallback(TemporaryCallbackSetter):
    """context manager for temporarily applying callbacks to the operational states of the provided components"""
    _state_set_attribute = 'operational_states'


def TemporaryCallback(*args, **kwargs):
    warnings.warn(  # v7.1.7
        'TemporaryCallback was renamed TemporaryOperationalCallback',
        DeprecationWarning,
        stacklevel=2,
    )
    return TemporaryOperationalCallback(*args, **kwargs)


class TemporaryActionCallback(TemporaryCallbackSetter):
    """context manager for temporarily applying callbacks to the action states of the provided components"""
    _state_set_attribute = 'action_states'


class ComponentState(InstanceRegistry):

    # dedicated component-specific mapping
    _components: MutableMapping[str, 'ComponentState'] = {}

    # override this when subclassing to indicate more advanced component action states
    action_states: Type[StateSet] = GenericActionState

    # override this when subclassing to indicate more advanced component operational states
    #   negative values indicate error states, 0 is offline, and positive is an operational state
    operational_states: Type[StateSet] = OperationalState

    def __init__(self,
                 component_name: str,
                 current_operational_state: Union[int, str, StateSet] = None,
                 current_action_state: Union[int, str, StateSet] = None
                 ):
        """
        ABC for tracking a component's operational state

        :param component_name: reference name for the component being registered
        :param current_operational_state: current state of the component
        """
        # todo create an InstanceRegistry subclass which tracks subclass instances (i.e. all components)
        if component_name in ComponentState.class_registry:
            raise NameError(
                f'the component "{component_name}" is already defined, please retrieve using get_component_by_name or '
                f'choose a different name for the component'
            )
        InstanceRegistry.__init__(self)
        ComponentState._components[component_name] = self
        self.component_name = component_name
        self.logger = state_logger.getChild(self.component_name)
        self._current_operational_state: OperationalState = None
        self.current_operational_state = current_operational_state or 0
        self._current_action_state: StateSet = None
        self.current_action_state = current_action_state or 0

        # set state names as parent IDs of callback list
        for state_set in [self.operational_states, self.action_states]:
            for state in state_set:
                state: StateSet
                if isinstance(state.value, State):
                    # set parent identifier override to be the string of the state
                    state.value.pre_state_callbacks.parent_identifier_override = state.__class__.__name__
                    state.value.post_state_callbacks.parent_identifier_override = state.__class__.__name__

    @staticmethod
    def _match_instance_key(inst: "ComponentState", identifier: str) -> bool:
        """matches the provided identifier against the component name and standard identifiers"""
        if inst._uuid == identifier:
            return True
        elif inst.uuid == identifier:
            return True
        elif identifier == inst.component_name:
            # todo consider matching case-insensitive
            return True
        return False

    @property
    def current_operational_state(self) -> StateSet:
        """The current state of the component"""
        return self._current_operational_state

    @current_operational_state.setter
    def current_operational_state(self, value: Union[int, str, StateSet]):
        self.update_operational_state(value)

    def _get_state_from_key(self, value: Union[int, str, StateSet], state_set: str) -> StateSet:
        """
        Gets an operational state from the provided key. If an StateSet instance which is not the same
        class as the assigned operational state class of the instance, the equivalent state will be retrieved.

        :param value: value, name, or StateSet
        :param state_set: state set key to use for retrieval
        """
        state_set = getattr(self, state_set)
        if isinstance(value, state_set):
            return value
        elif isinstance(value, StateSet):
            try:
                return state_set(value.value)
            except ValueError as e:
                raise ValueError(f'the operational state {value} does not have an equivalent {state_set.__name__} value')
        elif type(value) is str:
            return state_set[value.upper()]
        elif type(value) is int or isinstance(value, State):
            return state_set(value)
        else:
            raise TypeError(f'the type {type(value)} is not supported for a current component operational state')

    def get_operational_state_from_key(self, value: Union[int, str, StateSet]) -> StateSet:
        """
        Gets an operational state from the provided key. If an StateSet instance which is not the same
        class as the assigned operational state class of the instance, the equivalent state will be retrieved.

        :param value: value, name, or StateSet
        """
        return self._get_state_from_key(
            value,
            'operational_states',
        )

    def get_action_state_from_key(self, value: Union[int, str, StateSet]) -> StateSet:
        """
        Gets an action state from the provided key. If an StateSet instance which is not the same
        class as the assigned operational state class of the instance, the equivalent state will be retrieved.

        :param value: value, name, or StateSet
        """
        return self._get_state_from_key(
            value,
            'action_states',
        )

    def update_operational_state(self, state: Union[int, str, StateSet]) -> Optional[List]:
        """
        Updates the current state of the Component

        :param state: value to set the state to
        :return: return value from defined callback functions
        """
        state = self.get_operational_state_from_key(state)
        # execute callbacks if any are specified
        try:
            if state.has_pre_state_change_callbacks:
                self.logger.info(
                    f'executing {len(state.value.pre_state_callbacks)} pre-state-change callbacks for {state.name}'
                )
            out = state.execute_pre_state_callbacks()
        finally:
            self._current_operational_state = state
        if state.has_post_state_change_callbacks:
            self.logger.info(
                f'executing {len(state.value.pre_state_callbacks)} post-state-change callbacks for {state.name}'
            )
            out.extend(state.execute_post_state_callbacks())
        return out

    @property
    def current_action_state(self) -> StateSet:
        """the current action state of the component"""
        return self._current_action_state

    @current_action_state.setter
    def current_action_state(self, value: Union[int, str, StateSet]):
        self.update_action_state(value)

    def update_action_state(self, state: Union[int, str, StateSet]) -> Optional[List]:
        """
        Updates the current action state of the component

        :param state: value to set the state to
        :return: return value from defined callback functions
        """
        if type(state) is int or isinstance(state, State):
            state = self.action_states(state)
        elif type(state) is str:
            state = self.action_states[state.upper()]
        elif isinstance(state, self.action_states) is False:
            raise TypeError(f'the type {type(state)} is not supported for a current component action state')
        try:
            # execute callbacks if they are specified
            if state.has_pre_state_change_callbacks:
                self.logger.info(
                    f'executing {len(state.value.pre_state_callbacks)} pre-state-change callbacks for {state.name}'
                )
            out = state.execute_pre_state_callbacks()
        finally:
            self._current_action_state = state
        if state.has_post_state_change_callbacks:
            self.logger.info(
                f'executing {len(state.value.post_state_callbacks)} post-state-change callbacks for {state.name}'
            )
            out.extend(state.execute_post_state_callbacks())
        return out

    @classmethod
    def check_component_states(cls,
                               threshold_state: Union[int, OperationalState],
                               *components: Union[str, 'ComponentState'],
                               ) -> List['ComponentState']:
        """
        Checks the provided components against the minimum component state value, returning any components which do not
        pass the check.

        :param threshold_state: Minimum state value (components with states lower than this will fail the check)
        :param components: components to check
        :return: list of failing components
        """
        components = cls.get_components_to_monitor(*components)
        return [
            component
            for component in components
            if component.current_operational_state <= threshold_state
        ]

    @classmethod
    def check_component_states_raise(cls,
                                     threshold_state: Union[int, OperationalState],
                                     *components: 'ComponentState',
                                     ):
        """
        Checks the list of components against the minimum state and raises an error if any fail the test

        :param threshold_state: the threshold state to monitor for
        :param components: components to check
        """
        bad_states = cls.check_component_states(threshold_state, *components)
        if len(bad_states) > 0:
            state_logger.error(
                f'a bad state was encountered on the following component{"s" if len(bad_states) == 1 else ""}: '
                + ", ".join([
                    f'{component.component_name} {component.current_operational_state.value} '
                    f'({component.current_operational_state.name})'
                    for component in bad_states
                ])
            )
            raise BadComponentState(*bad_states, threshold_state=threshold_state)

    @classmethod
    def components_from_names(cls, *components: Union[str, 'ComponentState']) -> List['ComponentState']:
        """
        Retrieves a list of ComponentState subclasses from the provided list.

        :param components: components to find
        :return: list of ComponentState instances
        """
        out = []
        for component in components:
            if isinstance(component, ComponentState) is True:
                out.append(component)
            elif type(component) is str:
                out.append(cls._components[component])
            else:
                raise TypeError(f'the component "{component}" is not recognized')
        return out

    @classmethod
    def get_components_to_monitor(cls, *components: Union[str, 'ComponentState']) -> List['ComponentState']:
        """
        Gets a list of components to monitor from the provided arguments. If no components are provided, the complete
        list of components is provided.

        :param components: components to monitor
        :return: list of ComponentState instances
        """
        if len(components) > 0:
            return cls.components_from_names(*components)
        return [component for _, component in cls._components.items()]

    @classmethod
    def components_state_monitor_sleep(cls,
                                       duration: float,
                                       *components: Union[str, 'ComponentState'],
                                       cycle_time: float = 0.1,
                                       threshold_state: Union[int, StateSet] = OperationalState.OFFLINE,
                                       ):
        """
        Executes a sleep while also monitoring the state of the specified components. If no components are provided,
        all components will be monitored.

        :param duration: duration to sleep (approximate)
        :param components: components to monitor for a bad state. Can be either the name of the component or the
            ComponentState instance itself.
        :param cycle_time: cycle frequency (s; a check for a bad state will be executed every this seconds)
        :param threshold_state: the state to monitor for (states less or equal to this will raise an error)
        """
        # retrieve components as specified
        monitored_components = cls.get_components_to_monitor(*components)
        state_logger.info(
            f'waiting for ~{duration} s while monitoring '
            + f'{len(monitored_components)} components'
            if len(monitored_components) > 0
            else monitored_components[0].component_name
        )
        end_time = time.time() + duration
        while time.time() < end_time:
            cls.check_component_states_raise(threshold_state, *monitored_components)
            time.sleep(cycle_time)
        logging.info(f'finished waiting, continuing')

    def state_monitor_sleep(self,
                            duration: float,
                            cycle_time: float = 0.1,
                            threshold_state: Union[int, StateSet] = OperationalState.OFFLINE,
                            ):
        """
        Executes a sleep while also monitoring the state of the component. To monitor the state of multiple components,
        use components_state_monitor_sleep.

        :param duration: duration to sleep (approximate)
        :param cycle_time: cycle frequency (s; a check for a bad state will be executed every this seconds)
        :param threshold_state: the state to monitor for (states less or equal to this will raise an error)
        """
        return self.components_state_monitor_sleep(
            duration,
            self.component_name,
            cycle_time=cycle_time,
            threshold_state=threshold_state,
        )

    @classmethod
    def components_state_monitor_flag_check(cls,
                                            flag_check: Callable,
                                            *components: Union[str, 'ComponentState'],
                                            target_flag_state=True,
                                            cycle_time: float = 0.1,
                                            threshold_state: Union[int, StateSet] = OperationalState.OFFLINE,
                                            timeout: float = None,
                                            **flag_check_kwargs
                                            ):
        """
        Monitors the provided flag check while also monitoring the state of the specified components. If no components
        are provided, all components will be monitored.

        :param flag_check: method which will return a value that will be checked against the target_flag_state
        :param components: components to monitor for a bad state. Can be either the name of the component or the
            ComponentState instance itself.
        :param cycle_time: cycle frequency (s; a check for a bad state will be executed every this seconds)
        :param threshold_state: the state to monitor for (states less or equal to this will raise an error)
        :param target_flag_state: target flag state to monitor for
        :param timeout: optional timeout (seconds) to use while monitoring flags. If the timeout is reached before the
            flag state is reached, an error is raised.
        :param flag_check_kwargs: keyword arguments for the flag check
        """
        # retrieve components as specified
        monitored_components = cls.get_components_to_monitor(*components)
        state_logger.info(
            f'waiting for {flag_check.__name__} to be {target_flag_state} while monitoring '
            + f'{len(monitored_components)} components' if len(monitored_components) > 0
            else monitored_components[0].component_name
            + f' (timeout: {timeout} s)' if timeout is not None else ''
        )
        if timeout is not None:
            timeout = time.time() + timeout
        while flag_check(**flag_check_kwargs) != target_flag_state:
            if timeout is not None and time.time() > timeout:
                raise TimeoutError(
                    f'the desired flag state "{target_flag_state}" was not encountered before the timeout elapsed '
                )
            cls.check_component_states_raise(threshold_state, *monitored_components)
            time.sleep(cycle_time)
        logging.info(f'finished waiting, continuing')

    def state_monitor_flag_check(self,
                                 flag_check: Callable,
                                 target_flag_state=True,
                                 cycle_time: float = 0.1,
                                 threshold_state: Union[int, StateSet] = OperationalState.OFFLINE,
                                 timeout: float = None,
                                 **flag_check_kwargs
                                 ):
        """
        Monitors the state of the component, waiting for the threshold state. To monitor multiple components, use
        components_state_monitor_flag_check.


        :param flag_check: method which will return a value that will be checked against the target_flag_state
        :param cycle_time: cycle frequency (s; a check for a bad state will be executed every this seconds)
        :param threshold_state: the state to monitor for (states less or equal to this will raise an error)
        :param target_flag_state: target flag state to monitor for
        :param timeout: optional timeout (seconds) to use while monitoring flags. If the timeout is reached before the
            flag state is reached, an error is raised.
        :param flag_check_kwargs: keyword arguments for the flag check
        """
        self.components_state_monitor_flag_check(
            flag_check,
            self.component_name,
            target_flag_state=target_flag_state,
            cycle_time=cycle_time,
            threshold_state=threshold_state,
            timeout=timeout,
            **flag_check_kwargs,
        )

    @classmethod
    def get_component_by_name(cls, component_name: str) -> 'ComponentState':
        """
        Retrieves a component by its name. If you need to retrieve a component that you know is already instantiated,
        use this method.

        :param component_name: name assigned to the component
        """
        try:
            return cls._components[component_name]
        except KeyError:
            raise KeyError(f'the component "{component_name}" is not a defined component')

    @classmethod
    def get_or_create_component(cls, *args, component_name: str = None, **kwargs) -> 'ComponentState':
        """
        Retrieves an existing component or instantiates it as needed. This is intended to be the primary factory for
        component instantiation.

        :param args: component arguments
        :param component_name: name for the component
        :param kwargs: component keyword arguments
        :return: instantiated components
        """
        if component_name is None:
            raise ValueError(f'a component name must be specified')
        if component_name in cls._components:
            return cls._components[component_name]
        return cls(
            *args,
            component_name=component_name,
            **kwargs,
        )

    def wait_for_component_operational_state(self,
                                             desired_state: Union[int, StateSet, str],
                                             cycle_time: float = 0.1,
                                             timeout: float = None,
                                             comparison_operator: Callable = operator.ge,
                                             ):
        """
        Waits for the specified components to be at or above the specified operational state.

        :param desired_state: desired operational state for the component to be in
        :param cycle_time: state check cycle time
        :param timeout: optional timeout while waiting for the operational state
        :param comparison_operator: optional override for the comparison operator used by the wait method. The operator
            should expect two arguments: the current and desired states respectively. The operator function should return
            False if the condition is not met and True if it is.
        """
        # todo create a generic method to wait for multiple components
        desired_state = self.get_operational_state_from_key(desired_state)
        if timeout is not None:
            max_time = time.time() + timeout
        self.logger.info(f'waiting for {desired_state.name} operational state')
        while comparison_operator(self.current_operational_state, desired_state) is False:
            if timeout is not None and time.time() > max_time:
                msg = f'timeout of {timeout} s elapsed while waiting for {self.component_name} to be {desired_state.name}'
                self.logger.error(msg)
                raise TimeoutError(msg)
            time.sleep(cycle_time)

    def wait_for_component_action_state(self,
                                        desired_state: Union[int, StateSet, str],
                                        cycle_time: float = 0.1,
                                        timeout: float = None,
                                        comparison_operator: Callable = operator.ge,
                                        ):
        """
        Waits for the specified components to be at or above the specified action state.

        :param desired_state: desired action state for the component to be in
        :param cycle_time: state check cycle time
        :param comparison_operator: optional override for the comparison operator used by the wait method. The operator
            should expect two arguments: the current and desired states respectively. The operator function should return
            False if the condition is not met and True if it is.
        """
        # todo create a generic method to wait for multiple components
        desired_state = self.get_action_state_from_key(desired_state)
        if timeout is not None:
            max_time = time.time() + timeout
        self.logger.info(f'waiting for {desired_state.name} action state')
        while comparison_operator(self.current_action_state, desired_state) is False:
            if timeout is not None and time.time() > max_time:
                msg = f'timeout of {timeout} s elapsed while waiting for {self.component_name} to be {desired_state.name}'
                self.logger.error(msg)
                raise TimeoutError(msg)
            time.sleep(cycle_time)



"""functions for execution of specific actions"""
import logging

from .basic import Action
from .configured import ConfiguredAction, TrackedActionList, TrackedAction

logger = logging.getLogger(__name__)


@Action.register_action
def execute_sequence_actions(*actions: TrackedAction, **kwargs):
    """
    Executes a series of actions with the provided kwargs.

    e.g. if there is a method named "do_this_thing" which expects the keyword argument "value", the expected syntax
    of kwargs would be

        {
            "do_this_thing": {"value": 42},
        }

    :param actions: tracked action instances
    :param kwargs: additional keyword arguments for those instances.
    """
    if any([isinstance(action, TrackedAction) is False for action in actions]):
        raise TypeError('a non-TrackedAction object was provided as an action')
    for action in actions:
        action_kwargs = {}
        if action.configuration.name in kwargs:
            action_kwargs.update(kwargs[action.configuration.name])
        elif action.configuration.action.name in kwargs:
            action_kwargs.update(kwargs[action.configuration.action.name])
        logger.debug(f'executing {action}')
        action(**action_kwargs)
        if action.error is True:
            details = f'an error was encountered executing {action}: {action.error_details}'
            logger.error(details)
            raise RuntimeError(details)


@Action.register_action
def decision(*logic_checks: TrackedAction) -> ConfiguredAction:
    """
    Iterates through the provided Evaluation Actions until a True return is retrieved. The function will then return
    the next_action of that action.

    :param logic_checks: TrackedAction instances
    :return: next action of the first ConfiguredAction execution where True is determined
    """
    for action in logic_checks:
        # todo consider a try/except here to allow the provided actions to error without issue
        action()  # execute the action
        # if the action return is True, return the next action
        if action.action_return is True:
            return action.next_action


@Action.register_action
def run_automation(first_action: ConfiguredAction, **kwargs):
    """
    Runs an automation sequence by referencing the next action of each action after execution.

    :param first_action: first step to execute
    :param kwargs: keyword arguments corresponding to the actions
    """
    logger.info(f'beginning automation sequence with first step {first_action}')
    executed_actions: TrackedActionList = TrackedActionList()
    next_action = first_action
    while next_action is not None:
        # todo kwargs
        next_action = next_action.get_tracked_from_config()  # convert to tracked action
        logger.debug(f'executing {next_action}')
        if next_action.configuration.name in kwargs:
            logger.debug('updating kwargs as specified')
            next_action.update_kwargs(**kwargs[next_action.configuration.name])
        next_action()
        # append to list of executed actions
        executed_actions.append(next_action)
        if next_action.error is True:
            details = f'an error was encountered executing {next_action}: {next_action.error_details}'
            logger.error(details)
            logger.error(f'The executed actions before encountering the error: {executed_actions}')
            raise RuntimeError(details, executed_actions)
        # retrieve next action and continue
        next_action = next_action.next_action
    logger.info(f'automation completed, ran {len(executed_actions)} actions')
    return executed_actions
